# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
"""
Program Name: vl1l2_statistics.py
"""
import warnings
import numpy as np

from metcalcpy.util.met_stats import calc_speed
from metcalcpy.util.utils import round_half_up, sum_column_data_by_name, PRECISION, get_total_values

__author__ = 'Tatiana Burek'
__version__ = '0.1.0'


def calculate_vl1l2_bias(input_data, columns_names, aggregation=False):
    """Performs calculation of VL1L2_BIAS -

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated VL1L2_BIAS as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        uvffbar = sum_column_data_by_name(input_data, columns_names, 'uvffbar') / total
        uvoobar = sum_column_data_by_name(input_data, columns_names, 'uvoobar') / total
        bias = np.sqrt(uvffbar) - np.sqrt(uvoobar)
        result = round_half_up(bias, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_vl1l2_fvar(input_data, columns_names, aggregation=False):
    """Performs calculation of VL1L2_FVAR -

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated VL1L2_FVAR as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        uvffbar = sum_column_data_by_name(input_data, columns_names, 'uvffbar') / total
        f_speed_bar = sum_column_data_by_name(input_data, columns_names, 'f_speed_bar') / total
        result = uvffbar - f_speed_bar * f_speed_bar
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_vl1l2_ovar(input_data, columns_names, aggregation=False):
    """Performs calculation of VL1L2_OVAR -

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated VL1L2_OVAR as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        uvoobar = sum_column_data_by_name(input_data, columns_names, 'uvoobar') / total
        o_speed_bar = sum_column_data_by_name(input_data, columns_names, 'o_speed_bar') / total
        result = uvoobar - o_speed_bar * o_speed_bar
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_vl1l2_fspd(input_data, columns_names, aggregation=False):
    """Performs calculation of VL1L2_FSPD -

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated VL1L2_FSPD as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        ufbar = sum_column_data_by_name(input_data, columns_names, 'ufbar') / total
        vfbar = sum_column_data_by_name(input_data, columns_names, 'vfbar') / total
        fspd = calc_speed(ufbar, vfbar)
        result = round_half_up(fspd, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_vl1l2_ospd(input_data, columns_names, aggregation=False):
    """Performs calculation of VL1L2_OSPD -

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated VL1L2_OSPD as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        uobar = sum_column_data_by_name(input_data, columns_names, 'uobar') / total
        vobar = sum_column_data_by_name(input_data, columns_names, 'vobar') / total
        ospd = calc_speed(uobar, vobar)
        result = round_half_up(ospd, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_vl1l2_speed_err(input_data, columns_names, aggregation=False):
    """Performs calculation of VL1L2_SPEED_ERR -

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated VL1L2_SPEED_ERR as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        speed_bias = calculate_vl1l2_fspd(input_data, columns_names, aggregation) \
                     - calculate_vl1l2_ospd(input_data, columns_names, aggregation)
        result = round_half_up(speed_bias, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_vl1l2_msve(input_data, columns_names, aggregation=False):
    """Performs calculation of VL1L2_MSVE -

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated VL1L2_MSVE as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        uvffbar = sum_column_data_by_name(input_data, columns_names, 'uvffbar') / total
        uvfobar = sum_column_data_by_name(input_data, columns_names, 'uvfobar') / total
        uvoobar = sum_column_data_by_name(input_data, columns_names, 'uvoobar') / total
        msve = uvffbar - 2.0 * uvfobar + uvoobar
        if msve < 0:
            result = None
        else:
            result = round_half_up(msve, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_vl1l2_rmsve(input_data, columns_names, aggregation=False):
    """Performs calculation of VL1L2_RMSVE -
        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed
        Returns:
            calculated VL1L2_RMSVE as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        rmsve = np.sqrt(calculate_vl1l2_msve(input_data, columns_names, aggregation))
        result = round_half_up(rmsve, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_vl1l2_total(input_data, columns_names):
    """Performs calculation of Total number of matched pairs for
        Vector Partial Sums
        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated Total number of matched pairs as float
            or None if some of the data values are missing or invalid
    """
    total = sum_column_data_by_name(input_data, columns_names, 'total')
    return round_half_up(total, PRECISION)



# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
"""
Program Name: sum_stat.py

How to use:
 - Call from other Python function
        SUM_STAT = SumStat(PARAMS)
        SUM_STAT.calculate_stats_and_ci()
        where PARAMS – a dictionary with data description parameters including
        location of input and output data.
        The structure is similar to Rscript template

 - Run as a stand-alone script
        python sum_stat.py <parameters_file>
        where - <parameters_file> is YAML file with parameters
        and environment variable should be set to PYTHONPATH=<path_to_METcalcpy>

 - Run from Java
        proc = Runtime.getRuntime().exec(
                “python agg_stat.py <parameters_file>”,
                new String[]{”PYTHONPATH=<path_to_METcalcpy>”},
                new File(System.getProperty("user.home")));

"""
import sys
import itertools
import argparse
import time
import logging
import yaml
from inspect import signature

from metcalcpy import GROUP_SEPARATOR
from metcalcpy.util.ctc_statistics import *
from metcalcpy.util.grad_statistics import *
from metcalcpy.util.sl1l2_statistics import *
from metcalcpy.util.val1l2_statistics import *
from metcalcpy.util.vcnt_statistics import *
from metcalcpy.util.vl1l2_statistics import *
from metcalcpy.util.pstd_statistics import *
from metcalcpy.util.nbrcnt_statistics import *
from metcalcpy.util.ecnt_statistics import *
from metcalcpy.util.nbrctc_statistics import *
from metcalcpy.util.rps_statistics import *

from metcalcpy.util.utils import is_string_integer, parse_bool, \
    aggregate_field_values, perform_event_equalization, is_string_strictly_float


class SumStat:
    """A class that performs event equalisation if needed and statistics calculation
        on each row of the input data frame.
        if one of the field values contain ':' (EAST:NMT) this class aggregate the values
        for these fields first and than calculate statistics on the aggregated values
           All parameters including data description and location is in the parameters' dictionary
        Usage:
            initialise this call with the parameters dictionary and then
            call calculate_stats method
            This method will crate and save to the file aggregation statistics
                sum_stat = SumStat(params)
                sum_stat.calculate_stats()
        Raises: EmptyDataError or ValueError when the input DataFrame is empty or doesn't have data
       """

    def __init__(self, in_params):
        """Initialises the class by saving input parameters and reading data from file
            Args:
                in_params - input parameters as a dictionary
            Raises: EmptyDataError or ValueError when the input DataFrame is empty
                or doesn't have data
        """

        self.params = in_params
        # import pandas
        try:
            self.input_data = pd.read_csv(
                self.params['sum_stat_input'],
                header=[0],
                sep='\t'
            )

            if self.input_data.empty:
                logging.warning('The input data is empty. The empty output file was created')
                export_csv = self.input_data.to_csv(self.params['sum_stat_output'],
                                                    index=None, header=True, mode="w",
                                                    sep="\t")
                raise ValueError("The input data is empty")

            self.column_names = self.input_data.columns.values
        except pd.errors.EmptyDataError:
            raise
        except KeyError as ex:
            logging.error('Parameter with key %s is missing', ex)
            raise

    STATISTIC_TO_FIELDS = {'ctc': {"total", "fy_oy", "fy_on", "fn_oy", "fn_on"},
                           'sl1l2': {"total", "fbar", "obar", "fobar", "ffbar", "oobar"},
                           'grad': {"total", "fgbar", "ogbar", "mgbar", "egbar"},
                           'sal1l2': {"total", "fbar", "obar", "fobar", "ffbar", "oobar"},
                           'vl1l2': {"total", "ufbar", "vfbar", "uobar", "vobar", "uvfobar",
                                     "uvffbar", "uvoobar", "f_speed_bar", "o_speed_bar"},
                           'val1l2': {"total", "ufabar", "vfabar", "uoabar", "voabar",
                                      "uvfoabar", "uvffabar", "uvooabar"},
                           'pct': {'total', 'oy_i', 'on_i', 'thresh_i'},
                           'nbrcnt': {'total', 'fbs', 'fss', 'afss', 'ufss', 'f_rate', 'o_rate'},
                           'ecnt': {'total', 'rmse', 'rmse_oerr', 'crps'},
                           'nbrctc': {"total", "fy_oy", "fy_on", "fn_oy", "fn_on"},
                           'rps': {"total", "rpss", "rps", "rps_comp"}
                           }

    def calculate_stats(self):
        """ Calculates summary statistics for each series point
            Writes output data to the file
        """
        fields = []
        try:
            fields = self.STATISTIC_TO_FIELDS[self.params['line_type']]

            # change names for sal1l2 so we can reuse sl1l2 statistics
            if self.params['line_type'] == 'sal1l2':
                self.input_data = self.input_data.rename(columns={'fabar': 'fbar',
                                                                  'oabar': 'obar',
                                                                  'foabar': 'fobar',
                                                                  'ffabar': 'ffbar',
                                                                  'ooabar': 'oobar'
                                                                  })
                self.column_names = self.input_data.columns.values

            # perform EE if needed
            is_event_equal = parse_bool(self.params['event_equal'])
            if is_event_equal:
                self.input_data = perform_event_equalization(self.params, self.input_data)

            if not self.input_data.empty:
                # perform aggregation on a special field - needed for scorecard
                self.aggregate_special_fields('1')
                self.aggregate_special_fields('2')

                start_time = time.time()
                # calculate statistic for each row
                self.process_rows()
                # self.input_data = self.parallelize_on_rows(self.input_data,
                # self.process_row, num_of_processes=mp.cpu_count())
                print("--- %s seconds ---" % (time.time() - start_time))
            else:
                logging.warning('Event equalisation removed all data. '
                                'The empty output file is created')

        except KeyError as ex:
            logging.error('Parameter with key %s is missing. The empty output file is created', ex)

        # remove the original fields to save the space
        for column in fields:
            if column in self.column_names:
                self.input_data = self.input_data.drop(column, axis=1)

        # remove 'equalize' to save the space
        if 'equalize' in self.input_data.columns:
            self.input_data = self.input_data.drop(labels=['equalize'], axis=1)

        # save the result to file
        self.input_data.to_csv(self.params['sum_stat_output'],
                               index=None, header=True, mode='w',
                               sep="\t", na_rep="NA")

    def aggregate_special_fields(self, axis='1'):
        """Finds the data for special fields - the field with ';' (EAST;NMT) -
        and the requested axis in the initial dataframe
        and aggregate it based on the line type rules.
        If there are no special fields that require aggregation - does nothing
        The result is stored in the initial dataframe

        Args:
            axis - y1 or y1 axis
        """

        # check if indy_vals have a field that need to be aggregated - the field with ';'
        has_agg_indy_field = any(any(GROUP_SEPARATOR in i for i in item) for item in self.params['indy_vals'])

        # look if there is a field that need to be aggregated first - the field with ';'
        series_var_val = self.params['series_val_' + axis]
        has_agg_series_field = any(any(GROUP_SEPARATOR in i for i in item) for item in series_var_val)

        if series_var_val and (has_agg_indy_field or has_agg_series_field):
            # the special field was detected

            all_points = list(itertools.product(*series_var_val.values()))
            aggregated_values = pd.DataFrame()
            series_vars = list(series_var_val.keys())
            for indy_val in self.params['indy_vals']:
                # filter the input frame by each indy value
                if indy_val is None:
                    filtered_by_indy = self.input_data
                else:
                    # filter by value or split the value and filter by multiple values
                    filtered_by_indy = self.input_data[
                        self.input_data[self.params['indy_var']].isin(indy_val.split(';'))]

                for point in all_points:
                    point_data = filtered_by_indy

                    for index, series_var in enumerate(series_vars):
                        # get actual series value(s) and filter by them
                        if ';' in point[index]:
                            actual_series_vals = point[index].split(';')
                        else:
                            actual_series_vals = point[index].split(GROUP_SEPARATOR)
                        for ind, val in enumerate(actual_series_vals):
                            if is_string_integer(val):
                                actual_series_vals[ind] = int(val)
                            elif is_string_strictly_float(val):
                                actual_series_vals[ind] = float(val)
                        point_data = \
                            point_data[point_data[series_vars[index]].isin(actual_series_vals)]

                    # aggregate point data
                    if any(';' in series_val for series_val in point):
                        point_data = aggregate_field_values(series_var_val,
                                                            point_data,
                                                            self.params['line_type'])
                    elif ';' in indy_val:
                        # if aggregated value in indy val - add it to series values add aggregate
                        series_indy_var_val = series_var_val
                        series_indy_var_val[self.params['indy_var']] = [indy_val]
                        point_data = aggregate_field_values(series_indy_var_val,
                                                            point_data,
                                                            self.params['line_type'])

                    aggregated_values = aggregated_values.append(point_data)
            self.input_data = aggregated_values
            self.input_data.reset_index(inplace=True, drop=True)

    def process_rows(self):
        """For each row in the data frame finds the row statistic name,
            calculates it's value  and stores this value in the corresponding column
        """

        for index, row in self.input_data.iterrows():
            # statistic name
            stat = row['stat_name'].lower()

            # array of row's data
            row_array = np.expand_dims(row.to_numpy(), axis=0)

            # calculate the stat value
            stat_value = [calculate_statistic(row_array, self.column_names, stat)][0]

            # save the value to the 'stat_value' column
            self.input_data.at[index, 'stat_value'] = stat_value

    # def parallelize(self, data, func, num_of_processes=8):
    #     data_split = np.array_split(data, num_of_processes)
    #     pool = mp.Pool(num_of_processes)
    #     data = pd.concat(pool.map(func, data_split))
    #     pool.close()
    #     pool.join()
    #     return data
    #
    # def run_on_subset(self, func, data_subset):
    #     return data_subset.apply(func, axis=1)
    #
    # def parallelize_on_rows(self, data, func, num_of_processes=8):
    #     return self.parallelize(data, partial(self.run_on_subset, func), num_of_processes)


def calculate_statistic(values, columns_names, stat_name, aggregation=False):
    """Calculate the statistic of values
        Args:
            values: a np.array of values we want to calculate the statistic on
                    This is actually a 2d array (matrix) of values.
            stat_name: the name of the statistic
            aggregation: if the aggregation on fields was performed
        Returns:
            a  calculated statistics
        Raises:
            an error
        """
    func_name = f'calculate_{stat_name}'
    num_parameters = len(signature(globals()[func_name]).parameters)
    if num_parameters == 2:
        stat = globals()[func_name](values, columns_names)
    else:
        stat = globals()[func_name](values, columns_names, aggregation)
    return stat


if __name__ == "__main__":
    PARSER = argparse.ArgumentParser(description='List of sum_stat arguments')
    PARSER.add_argument("parameters_file", help="Path to YAML parameters file",
                        type=argparse.FileType('r'),
                        default=sys.stdin)
    ARGS = PARSER.parse_args()
    PARAMS = yaml.load(ARGS.parameters_file, Loader=yaml.FullLoader)

    SUM_STAT = SumStat(PARAMS)
    SUM_STAT.calculate_stats()



# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
import os
import numpy as np
import xarray as xr
import datetime
import pandas as pd
from scipy.fftpack import rfft, irfft, fftfreq
from scipy.signal import detrend

def rmm(olr, u850, u200, time, spd, EOF1, EOF2, rmm_norm, pc_norm):
    """
    Compute RMM index for given olr, u850 and u200 averaged from 15S - 15N. Use observed RMM EOFs 
    to project the data onto. To match the observed RMM index use ERA Interim wind and observed OLR 
    anomalies from 1979-2001 climatology. The normalization factors are the original ones from Wheeler 
    and Hendon (2004).
    :param olr: OLR data (time, lon) DataArray
    :param u850: zonal wind at 850hPa (time, lon) DataArray
    :param u200: zonal wind at 200hPa (time, lon) DataArray
    :param time: time datetime64 array
    :param spd: number of obs per day
    :parm rmm_norm: normalization factors for OLR, U850, and U200 from 1979-2001
    :parm pc_norm: normalization factors for PCs from 1979-2001
    :return: RMM PCs
    """



    # test that the input data has the same longitude values as the EOFs
    eoflon = EOF1['lon']
    lon = olr['lon']
    if not np.all(lon == eoflon):
        raise ValueError("Longitude grid of EOFs and OLR is not equal.  Longitudes should be 0 to 375.5 by 2.5 degrees.")
    lon = u850['lon']
    if not np.all(lon == eoflon):
        raise ValueError("Longitude grid of EOFs and U850 is not equal.  Longitudes should be 0 to 375.5 by 2.5 degrees.")
    lon = u200['lon']
    if not np.all(lon == eoflon):
        raise ValueError("Longitude grid of EOFs and U200 is not equal.  Longitudes should be 0 to 375.5 by 2.5 degrees.")

    # remove previous 120 day mean from data
    olr = rem_120day_mean(olr,time)
    u850 = rem_120day_mean(u850,time)
    u200 = rem_120day_mean(u200,time)

    # normalize by the square root of the average variance from 1979-2001
    olr.values   = olr/rmm_norm[0] 
    u850.values  = u850/rmm_norm[1]
    u200.values  = u200/rmm_norm[2]

    # add all variables to a single array (append in longitude)
    dims = olr.shape
    ntim = dims[0]
    nlon = dims[1]

    data = xr.DataArray(np.empty([ntim,3*nlon]), dims=('time','lon_ext'))
    data[:,0:nlon] = olr
    data[:,nlon:2*nlon] = u850 
    data[:,2*nlon::] = u200

    # add EOFs to single array for all variables
    eof1 = xr.DataArray(np.empty([3*nlon]), dims=('lon_ext'))
    eof2 = xr.DataArray(np.empty([3*nlon]), dims=('lon_ext'))
    for i in np.arange(0,3):
        eof1[nlon*(i):nlon*(i+1)] = EOF1[i,:]
        eof2[nlon*(i):nlon*(i+1)] = EOF2[i,:]

    # project data onto the EOF patterns
    raw_pc1, raw_pc2 = regress_2dim_data_onto_eofs(data, time, eof1, eof2)

    # normalize PCs by their std from 1979 - 2001
    normalization_factor = 1 / pc_norm[0]
    #normalization_factor = 1 / np.std(raw_pc1)
    pc1 = np.multiply(raw_pc1, normalization_factor)
    normalization_factor = 1 / pc_norm[1]
    #normalization_factor = 1 / np.std(raw_pc2)
    pc2 = np.multiply(raw_pc2, normalization_factor)

    pc1 = xr.DataArray(pc1,dims=['time'],coords={'time':time})
    pc2 = xr.DataArray(pc2,dims=['time'],coords={'time':time})

    return pc1, pc2


def omi(olr, time, spd, EOF1, EOF2):
    """
    Compute OMI index for given input OLR. Use observed OMI EOFs to project the data onto. To reproduce
    the observed OMI use daily OLR from PSL/NOAA at https://psl.noaa.gov/data/gridded/data.interp_OLR.html
    and the filtering method perform_3dim_spectral_filtering. This is slower than using the bandpass_filter_3D
    filter method, but it matches the observed OMI values closely. 

    :param olr: OLR data  (time, lat, lon) DataArray
    :param time: datetime64 time array
    :param spd: number of obs per day
    :param EOF1: filepath to the location of the eof1 files
    :param EOF2: filepath to the location of the eof2 files
    :return: OMI PCs
    """
    #EOF1, EOF2 = read_omi_eofs(eof1_path, eof2_path)

    # check that all latitudes and longitudes match between EOFs and OLR
    eoflon = EOF1['lon']
    lon = olr['lon']
    eoflat = EOF1['lat']
    lat = olr['lat']
    if not np.all(lat == eoflat):
        raise ValueError("Latitude grid of EOFs and OLR is not equal. Latitudes should be -20 to 20 by 2.5 degrees.")
    if not np.all(lon == eoflon):
        raise ValueError("Longitude grid of EOFs and OLR is not equal. Longitudes should be 0 to 375.5 by 2.5 degrees.")

    # filter OLR for 20-96 days
    #filtered_olr_data  = bandpass_filter_3D(olr, time, 20, 96)
    time_spacing = (time[1] - time[0]).astype('timedelta64[s]') / np.timedelta64(1, 'D')  # time spacing in days
    filtered_olr_data = perform_3dim_spectral_filtering(olr, time_spacing, 20, 96, -720, 720)

    # regress OLR onto daily EOFs
    raw_pc1, raw_pc2 = regress_3dim_data_onto_eofs(filtered_olr_data, time, EOF1, EOF2)

    # normalize PCs
    normalization_factor = 1 / np.nanstd(raw_pc1)
    pc1 = np.multiply(raw_pc1, normalization_factor)
    pc2 = np.multiply(raw_pc2, normalization_factor)

    pc1 = xr.DataArray(pc1,dims=['time'],coords={'time':time})
    pc2 = xr.DataArray(pc2,dims=['time'],coords={'time':time})

    return pc1, pc2



def regress_3dim_data_onto_eofs(data, time, EOF1, EOF2):
    """
    Finds time-dependent coefficients w.r.t the DOY-dependent EOF basis for time-dependent spatially 
    resolved data.
    I.e. it finds the PCs for temporally resolved OLR data.
    :param data: data DataArray to project onto the EOFs
    :param time: datetime64 time array
    :param EOF1: the day of year dependent EOF1
    :param EOF2: the day of year dependent EOF2
    :return: PCs
    """
    if not np.all(data.lat == EOF1.lat):
        raise ValueError("Latitude grid of EOFs and OLR is not equal.")
    if not np.all(data.lon == EOF1.lon):
        raise ValueError("Longitude grid of EOFs and OLR is not equal.")
    pc1 = np.empty(data.time.size)
    pc2 = np.empty(data.time.size)
    nlat = len(data.lat)
    nlon = len(data.lon)

    for idx, val in enumerate(time):
        olr_singleday = data.sel(time=val)
        # get doy, only works of day is datetime or date object
        doy = pd.to_datetime(val).timetuple().tm_yday
        
        (pc1_single, pc2_single) = regress_vector_onto_eofs(
            np.reshape(olr_singleday.values,(nlat*nlon)), np.reshape(EOF1[doy-1,:,:].values,(nlat*nlon)), 
            np.reshape(EOF2[doy-1,:,:].values,(nlat*nlon)) )
        
        pc1[idx] = pc1_single
        pc2[idx] = pc2_single

    return pc1, pc2


def regress_2dim_data_onto_eofs(data, time, EOF1, EOF2):
    """
    Finds time-dependent coefficients w.r.t the 2D EOF basis for time-dependent longitudinally 
    resolved data.
    I.e. it finds the PCs for temporally resolved OLR data.
    :param data: data DataArray to project onto the EOFs
    :param time: datetime64 time array
    :param EOF1: the day of year dependent EOF1
    :param EOF2: the day of year dependent EOF2
    :return: PCs
    """
    pc1 = np.empty(time.size)
    pc2 = np.empty(time.size)
    nlon = len(data.lon_ext)

    for idx, val in enumerate(time):
        olr_singleday = data[idx,:]
        
        (pc1_single, pc2_single) = regress_vector_onto_eofs(
            np.reshape(olr_singleday.values,(nlon)), np.reshape(EOF1.values,(nlon)), 
            np.reshape(EOF2.values,(nlon)) )

        pc1[idx] = pc1_single
        pc2[idx] = pc2_single

    return pc1, pc2    


def regress_vector_onto_eofs(vector: np.ndarray, eof1: np.ndarray, eof2: np.ndarray):
    """
    Helper method that finds the coefficients of the given vector with respect to the given basis of 
    2 EOFs. The computed coefficients are the PCs in the terminology of the EOF analysis.
    :param vector: The vector for which the coefficients in the EOF basis should be found.
    :param eof1: EOF basis vector 1.
    :param eof2: EOF basis vector 2.
    :return: The two PCs.
    """
    eof_mat = np.array([eof1, eof2]).T

    # Alternative implementation 1:
    x = np.linalg.lstsq(eof_mat, vector, rcond=None)
    pc1, pc2 = x[0]

    return pc1, pc2


def bandpass_filter_3D(data, time, a, b):
    """
    Filter 3D data in time. Bandpass filter limits are given in units of days. Assumes filter dimension 
    is the first dimension of the array (time x lat x lon). This method is much faster than the spectral
    filter in perform_3dim_spectral_filtering below. However, it does not give the same answer! If speed
    is of importance this routine can be used, but for accurancy  perform_3dim_spectral_filtering is
    reccomended.
    :param data: input DataArray to be filtered in time (time, lat, lon)
    :param time: datetime64 time array
    :param a: short period limit in days
    :param b: long period limit in days
    :return: filtered DataArray the same size as the input array
    """
    filtered_data = data.copy()

    dt = (time[1] - time[0]).astype('timedelta64[s]') / np.timedelta64(1, 'D')

    # remove linear trend from data in time
    data = detrend(data,axis=0,type='linear',overwrite_data=True)

    # taper time window and pad with zeros
    orig_data = data
    orig_nt, nm, nl = orig_data.shape
    
    nt = 2 ** 17  # Zero padding for performance and resolution optimization, as well as consistency with origininal Kiladis code

    if orig_nt > nt:
        raise ValueError('Time series is longer than hard-coded value for zero-padding!')

    data = np.zeros([nt, nm, nl])
    data[0:orig_nt, :, :] = orig_data

    # 10 days tapering according ot Kiladis Code
    # only relevant at beginning of time series as it is zero-padded in the end
    for idx_l in range(0, nl):
        for idx_m in range(0, nm):
            data[:, idx_m, idx_l] = taper_vector_to_zero(data[:, idx_m, idx_l], int(10 / dt))

    # set up frequency array for masking

    W = fftfreq(data[:,0,0].size, dt)
    W = np.tile(W,(nm,nl,1))
    W = np.moveaxis(W,2,0)
    P = 1 / W

    # run fft
    data_fft = rfft(data,axis=0)
        
    data_fft = np.where(P<a, 0, data_fft)
    data_fft = np.where(P>b, 0, data_fft)

    filter = irfft(data_fft,axis=0)
    filtered_data.values = filter[0:orig_nt, :, :]

    return filtered_data  


def rem_120day_mean(data, time):
    """
    Filter 3D data in time. Bandpass filter limits are given in units of days. Assumes filter dimension 
    is the first dimension of the array (time x lat x lon)
    :param data: input DataArray to remove previous 120 day time mean from
    :param time: datetime64 time array
    :return: DataArray containing the data with the previous 120day mean removed from each time step
    """
    filtered_data = data.copy()
    dt = (time[1] - time[0]).astype('timedelta64[s]') / np.timedelta64(1, 'D')
    ntim_mean = int(120/dt)

    for idx, val in enumerate(time):
        if idx>ntim_mean:
            filtered_data[idx,:] = filtered_data[idx,:].values - np.nanmean(data[idx-ntim_mean:idx+1,:].values, axis=0)
        else:
            filtered_data[idx,:] = filtered_data[idx,:].values - np.nanmean(data[0:idx+1,:].values, axis=0)   

    return filtered_data       


def zon_avg_variance(data):
    """
    Compute zonally averaged temporal variance for 2D data.
    :param data:  DataArray (time x lon) 
    :return: numpy array (1)
    """
    # compute temporal variance at each longitude
    var = np.nanvar(data,axis=0)

    # compute zonal mean of the temporal variance
    zavg_var  = np.nanmean(var)     

    return zavg_var     


def taper_vector_to_zero(data, window_length):
    """
    Taper the data in the given vector to zero at both the beginning and the ending.
    :param data: The data to taper.
    :param window_length: The length of the window (measured in vector indices),
        in which the tapering is applied for the beginning and the ending independently
    :return: The tapered data.
    """
    startinds = np.arange(0, window_length, 1)
    endinds = np.arange(-window_length - 1, -1, 1) + 2

    result = data
    result[0:window_length] = result[0:window_length] * 0.5 * (1 - np.cos(startinds * np.pi / window_length))
    result[data.size - window_length:data.size] = \
        result[data.size - window_length:data.size] * 0.5 * (1 - np.cos(endinds * np.pi / window_length))
    return result


def perform_3dim_spectral_filtering(data, time_spacing, period_min, period_max, 
    wn_min: float, wn_max: float):
        """
        Bandpass-filters OLR data in time- and longitude-direction according to
        the original Kiladis algorithm.
        Note that the temporal and longitudinal dimension have in principle
        different characteristics, so that they are in detail treated a bit
        differently:
        While the time is evolving into infinity (so that the number of data
        points and the time_spacing variable are needed to calculate the
        full temporal coverage), the longitude is periodic with the periodicity
        of one globe (so that it is assumed that the data covers exactly one
        globe and only passing the number of longitudes provides already the complete information).
        :param data: The OLR data as 2-dim array: first dimension time, second
            dimension longitude, both equally spaced.
            The longitudinal dimension has to cover the full globe.
            The time dimension is further described by the variable
            `time_spacing`.
        :param time_spacing: Temporal resolution of the data in days (often 1 or 0.5 (if two
            data points exist per day)).
        :param period_min: Minimal period (in days) that remains in the dataset.
        :param period_max: Maximal period (in days) that remains in the dataset.
        :param wn_min: Minimal wavenumber (in cycles per globe) that remains in the dataset.
        :param wn_max: Maximal wavenumber (in cycles per globe) that remains in the dataset.
        :param do_plot: If True, diagnosis plots will be generated.
        :param save_debug: If true, some variables will be filled with values of intermediate processing steps
            for debugging purposes.
        :return: The filtered data.
        """

        # ###################### Process input data #######################
        dataperday = 1 / time_spacing
        freq_min = 1 / period_max
        freq_max = 1 / period_min

        # ######################## Detrend #################################
        # "orig" refers to the original size in the time dimension in the following, i.e. not the zero-padded version.
        filtered_data = data.copy()
        orig_data = data
        orig_nt, nm, nl = orig_data.shape
        lat = orig_data['lat']

        print('detrending in time')
        orig_data = detrend(orig_data, axis=0, type='linear', overwrite_data=True)

        # ######################## Zero Padding ############################
        nt_max = 2 ** 17  # Zero padding for performance and resolution optimization, as well as consistency with origininal Kiladis code
        #nt = nt_max
        nt = orig_nt+1000 # this is a much faster option, but has slight differences to the original OMI

        if orig_nt > nt_max:
            raise ValueError('Time series is longer than hard-coded value for zero-padding!')

        data = np.zeros([nt, nm, nl])
        data[0:orig_nt, :, :] = orig_data

        # ######################## Tapering to zero ########################
        # 10 days tapering according ot Kiladis Code
        # only relevant at beginning of time series as it is zero-padded in the end
        print('taper beginning of time series to zero')
        for idx_l in range(0, nl):
            for idx_m in range(0, nm):
                data[:, idx_m, idx_l] = taper_vector_to_zero(data[:, idx_m, idx_l], int(10 * dataperday))

        # ########################## Forward Fourier transform ############
        

        # ## Calculation of the frequency grid in accordance with Kiladis code
        freq_axis = np.zeros(nt)
        for i_f in range(0, nt):
            if (i_f <= nt / 2):
                freq_axis[i_f] = i_f * dataperday / nt
            else:
                freq_axis[i_f] = -1 * (nt - i_f) * dataperday / nt
        # the following code based on scipy function produces qualitatively the same grid.
        # However, numerical differences seem to have larger effect for the filtering step.
        # freq_axis = np.fft.fftfreq(nt, d=time_spacing)
        # freq_axis = np.fft.fftshift(freq_axis)
        # freq_axis = np.roll(freq_axis, int(nt/2))

        # ## Calculation of the wavenumber grid in accordance with Kiladis code
        wn_axis = np.zeros(nl)
        for i_wn in range(0, nl):
            if i_wn <= nl / 2:
                wn_axis[i_wn] = -1 * i_wn
                # note: to have this consistent with the time-dimension, one could write wn_axis[i_wn]= -1*i_wn*dataperglobe/nl
                # However, since data is required to cover always one globe nl will always be equal to dataperglobe
                # The sign is not consistent with the time dimension, which is for reasons of consitency with the original Kiladis implementation
            else:
                wn_axis[i_wn] = nl - i_wn
        # the following code based on scipy function produces qualitatively the same grid.
        # However, numerical differences seem to have larger effect for the filtering step.
        # wn_axis = np.fft.fftfreq(nl, d=dy)
        # wn_axis = np.fft.fftshift(wn_axis)  #identical with  wn_axis=np.arange(-int(nlong/2), int(nlong/2),1.)
        # wn_axis = -1 *wn_axis
        # wn_axis = np.roll(wn_axis, int(nl/2))

        # ################### Filtering of the Fourier Spectrum #############
        # ## name filter boundaries like in Kiladis Fortran Code
        f1 = freq_min
        f2 = freq_min
        f3 = freq_max
        f4 = freq_max
        s1 = wn_min
        s2 = wn_max
        s3 = wn_min
        s4 = wn_max

        for idx_m, ilat in enumerate(lat):
            print('filtering latitude: '+str(ilat))
            fourier_fft = np.fft.fft2(np.squeeze(data[:,idx_m,:]))
            # reordering of spectrum is done to be consistent with the original kiladis ordering.
            fourier_fft = np.fft.fftshift(fourier_fft, axes=(0, 1))
            fourier_fft = np.roll(fourier_fft, int(nt / 2), axis=0)
            fourier_fft = np.roll(fourier_fft, int(nl / 2), axis=1)

            # ### Very similar to original Kiladis Code
            fourier_fft_filtered = fourier_fft
            count = 0
            for i_f in range(0, int(nt / 2) + 1):
                for i_wn in range(0, nl):
                    ff = freq_axis[i_f]
                    ss = wn_axis[i_wn]
                    if ((ff >= ((ss * (f1 - f2) + f2 * s1 - f1 * s2) / (s1 - s2))) and
                            (ff <= ((ss * (f3 - f4) + f4 * s3 - f3 * s4) / (s3 - s4))) and
                            (ss >= ((ff * (s3 - s1) - f1 * s3 + f3 * s1) / (f3 - f1))) and
                            (ss <= ((ff * (s4 - s2) - f2 * s4 + f4 * s2) / (f4 - f2)))):
                        count = count + 1
                    else:
                        fourier_fft_filtered[i_f, i_wn] = 0
                        if i_wn == 0 and i_f == 0:
                            pass
                        elif i_wn == 0:
                            ind_f = nt - i_f
                            if (ind_f < nt):
                                fourier_fft_filtered[ind_f, i_wn] = 0
                        elif i_f == 0:
                            ind_wn = nl - i_wn
                            if ind_wn < nl:
                                fourier_fft_filtered[i_f, ind_wn] = 0
                        else:
                            ind_f = nt - i_f
                            ind_wn = nl - i_wn
                            if ind_f < nt and ind_wn < nl:
                                fourier_fft_filtered[ind_f, ind_wn] = 0

            # ############################ FFT Backward transformation ############
            # #reorder spectrum back from kiladis ordering to python ordering
            fourier_fft_filtered = np.roll(fourier_fft_filtered, -int(nt / 2), axis=0)
            fourier_fft_filtered = np.roll(fourier_fft_filtered, -int(nl / 2), axis=1)
            fourier_fft_filtered = np.fft.ifftshift(fourier_fft_filtered, axes=(0, 1))
            filtered_olr = np.fft.ifft2(fourier_fft_filtered)
            filtered_olr = np.real(filtered_olr)

            # ############################# remove zero padding elements ##########
            filtered_data[:, idx_m, :] = filtered_olr[0:orig_nt, :] 

        # ToDo: Make sure that result is real
        return filtered_data



# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
"""
Program Name: agg_stat.py

How to use:
 - Call from other Python function
        AGG_STAT = AggStat(PARAMS)
        AGG_STAT.calculate_stats_and_ci()
        where PARAMS – a dictionary with data description parameters including
        location of input and output data.
        The structure is similar to Rscript template

 - Run as a stand-alone script
        python agg_stat.py <parameters_file>
        where - <parameters_file> is YAML file with parameters
        and environment variable should be set to PYTHONPATH=<path_to_METcalcpy>

 - Run from Java
        proc = Runtime.getRuntime().exec(
                “python agg_stat.py <parameters_file>”,
                new String[]{”PYTHONPATH=<path_to_METcalcpy>”},
                new File(System.getProperty("user.home")));

"""
import sys
import itertools
import argparse
from inspect import signature
import yaml
import pandas

from metcalcpy import GROUP_SEPARATOR, DATE_TIME_REGEX
from metcalcpy.bootstrap import bootstrap_and_value, BootstrapResults
from metcalcpy.util.ctc_statistics import *
from metcalcpy.util.grad_statistics import *
from metcalcpy.util.sl1l2_statistics import *
from metcalcpy.util.sal1l2_statistics import *
from metcalcpy.util.ssvar_statistics import *
from metcalcpy.util.val1l2_statistics import *
from metcalcpy.util.vcnt_statistics import *
from metcalcpy.util.vl1l2_statistics import *
from metcalcpy.util.ecnt_statistics import *
from metcalcpy.util.nbrcnt_statistics import *
from metcalcpy.util.nbrctc_statistics import *
from metcalcpy.util.pstd_statistics import *
from metcalcpy.util.rps_statistics import *
from metcalcpy.util.mcts_statistics import *

from metcalcpy.util.utils import is_string_integer, get_derived_curve_name, \
    calc_derived_curve_value, intersection, is_derived_point, parse_bool, \
    OPERATION_TO_SIGN, perfect_score_adjustment, perform_event_equalization, \
    aggregate_field_values, sort_data, DerivedCurveComponent, is_string_strictly_float

__author__ = 'Tatiana Burek'


class AggStat:
    """A class that performs aggregation statistic logic on input data frame.
           All parameters including data description and location is in the parameters dictionary
           Usage:
                initialise this call with the parameters dictionary and than
                call calculate_stats_and_ci method
                This method will crate and save to the file aggregation statistics
                    agg_stat = AggStat(params)
                    agg_stat.calculate_stats_and_ci()
            Raises: EmptyDataError or ValueError when the input DataFrame is empty
                or doesn't have data
       """

    def __init__(self, in_params):
        """Initialises the class by saving input parameters and reading data from file

            Args:
                in_params - input parameters as a dictionary
            Raises: EmptyDataError or ValueError when the input DataFrame is empty
                or doesn't have data
        """

        self.statistic = None
        self.derived_name_to_values = {}
        self.params = in_params

        try:
            self.input_data = pd.read_csv(
                self.params['agg_stat_input'],
                header=[0],
                sep='\t'
            )
            self.column_names = self.input_data.columns.values
        except pandas.errors.EmptyDataError:
            raise
        except KeyError as er:
            print(f'ERROR: parameter with key {er} is missing')
            raise
        self.group_to_value = {}

    EXEMPTED_VARS = ['SSVAR_Spread', 'SSVAR_RMSE']
    STATISTIC_TO_FIELDS = {
        'baser': ['fy_oy', 'fn_oy'],
        'acc': ['fy_oy', 'fn_on'],
        'fbias': ['fy_oy', 'fn_on', 'fy_oy', 'fy_on'],
        'fmean': ['fy_oy', 'fy_on'],
        'pody': ['fy_oy', 'fn_oy'],
        'pofd': ['fy_on', 'fn_on'],
        'podn': ['fn_on', 'fy_on'],
        'far': ['fy_on', 'fy_oy'],
        'csi': ['fy_oy', 'fy_on', 'fn_oy'],
        'gss': ['fy_oy', 'fy_on', 'fn_oy'],
        'hk': ['fy_oy', 'fn_oy', 'fy_on', 'fn_on'],
        'hss': ['fy_oy', 'fn_oy', 'fy_on', 'fn_on'],
        'odds': ['fy_oy', 'fn_oy', 'fy_on', 'fn_on'],
        'lodds': ['fy_oy', 'fn_oy', 'fy_on', 'fn_on'],
        'baggs': ['fy_oy', 'fn_oy', 'fy_on'],
        'eclv': ['fy_oy', 'fn_oy', 'fy_on', 'fn_on'],

        'fbar': ['fbar'],
        'obar': ['obar'],
        'fstdev': ['fbar', 'ffbar'],
        'ostdev': ['obar', 'oobar'],
        'fobar': ['fobar'],
        'ffbar': ['ffbar'],
        'oobar': ['oobar'],
        'mae': ['mae'],
        'mbias': ['obar', 'fbar'],
        'corr': ['ffbar', 'fbar', 'oobar', 'obar', 'fobar'],
        'anom_corr': ['ffabar', 'fabar', 'ooabar', 'oabar', 'foabar'],
        'anom_corr_raw': ['ffabar', 'ooabar', 'foabar'],
        'rmsfa': ['ffabar'],
        'rmsoa': ['ooabar'],
        'me': ['fbar', 'obar'],
        'me2': ['fbar', 'obar'],
        'mse': ['ffbar', 'oobar', 'fobar'],
        'msess': ['ffbar', 'oobar', 'fobar', 'obar'],
        'rmse': ['ffbar', 'oobar', 'fobar'],
        'si': ['ffbar', 'oobar', 'fobar', 'obar'],
        'estdev': ['ffbar', 'oobar', 'fobar', 'fbar', 'obar'],
        'bcmse': ['ffbar', 'oobar', 'fobar', 'fbar', 'obar'],
        'bcrmse': ['ffbar', 'oobar', 'fobar', 'fbar', 'obar'],
        'pr_corr': ['ffbar', 'oobar', 'fobar', 'fbar', 'obar'],

        'fgbar': ['fgbar'],
        'ogbar': ['ogbar'],
        'mgbar': ['mgbar'],
        'egbar': ['egbar'],
        's1': ['egbar', 'mgbar'],
        's1_og': ['egbar', 'ogbar'],
        'fgog_ratio': ['fgbar', 'ogbar'],

        'vcnt_fbar': ['f_speed_bar'],
        'vcnt_obar': ['o_speed_bar'],
        'vcnt_fs_rms': ['uvffbar'],
        'vcnt_os_rms': ['uvoobar'],
        'vcnt_msve': ['uvffbar', 'uvfobar', 'uvoobar'],
        'vcnt_rmsve': ['uvffbar', 'uvfobar', 'uvoobar'],
        'vcnt_fstdev': ['uvffbar', 'f_speed_bar'],
        'vcnt_ostdev': ['uvoobar', 'o_speed_bar'],
        'vcnt_fdir': ['ufbar', 'vfbar'],
        'vcnt_odir': ['uobar', 'vobar'],
        'vcnt_fbar_speed': ['ufbar', 'vfbar'],
        'vcnt_obar_speed': ['uobar', 'vobar'],
        'vcnt_vdiff_speed': ['ufbar', 'uobar', 'vfbar', 'vobar'],
        'vcnt_vdiff_dir': ['ufbar', 'uobar', 'vfbar', 'vobar'],
        'vcnt_speed_err': ['ufbar', 'vfbar', 'uobar', 'vobar'],
        'vcnt_speed_abserr': ['ufbar', 'vfbar', 'uobar', 'vobar'],
        'vcnt_dir_err': ['ufbar', 'vfbar', 'uobar', 'vobar'],
        'vcnt_dir_abser': ['ufbar', 'vfbar', 'uobar', 'vobar'],
        'vcnt_anom_corr': ['uvffabar', 'uvfoabar', 'uvooabar', 'fa_speed_bar', 'oa_speed_bar'],
        'vcnt_anom_corr_uncntr': ['uvffabar', 'uvfoabar', 'uvooabar'],

        'vl1l2_bias': ['uvffbar', 'uvoobar'],
        'vl1l2_fvar': ['uvffbar', 'f_speed_bar'],
        'vl1l2_ovar': ['uvoobar', 'o_speed_bar'],
        'vl1l2_speed_err': ['ufbar', 'vfbar', 'uobar', 'vobar'],
        'vl1l2_rmsve': ['uvffbar', 'uvfobar', 'uvoobar'],
        'vl1l2_msve': ['uvffbar', 'uvfobar', 'uvoobar'],

        'val1l2_anom_corr':
            ['ufabar', 'vfabar', 'uoabar', 'voabar', 'uvfoabar', 'uvffabar', 'uvooabar'],

        'ssvar_fbar': ['fbar'],
        'ssvar_fstdev': ['fbar', 'ffbar'],
        'ssvar_obar': ['obar'],
        'ssvar_ostdev': ['obar', 'oobar'],
        'ssvar_pr_corr': ['ffbar', 'fbar', 'oobar', 'obar', 'fobar'],
        'ssvar_me': ['fbar', 'obar'],
        'ssvar_estdev': ['obar', 'fbar', 'ffbar', 'oobar', 'fobar'],
        'ssvar_mse': ['ffbar', 'oobar', 'fobar'],
        'ssvar_bcmse': ['fbar', 'obar', 'ffbar', 'oobar', 'fobar'],
        'ssvar_bcrmse': ['fbar', 'obar', 'ffbar', 'oobar', 'fobar'],
        'ssvar_rmse': ['ffbar', 'oobar', 'fobar'],
        'ssvar_anom_corr': ['fbar', 'obar', 'ffbar', 'oobar', 'fobar'],
        'ssvar_me2': ['fbar', 'obar'],
        'ssvar_msess': ['obar', 'oobar', 'ffbar', 'fobar'],
        'ssvar_spread': ['var_mean'],

        'ecnt_crps': ['crps'],
        'ecnt_crpss': ['crps', 'crpscl'],
        'ecnt_ign': ['ign'],
        'ecnt_me': ['me'],
        'ecnt_rmse': [],
        'ecnt_spread': ['spread'],
        'ecnt_me_oerr': ['me_oerr'],
        'ecnt_rmse_oerr': [],
        'ecnt_spread_oerr': ['spread_oerr'],
        'ecnt_spread_plus_oerr': ['spread_plus_oerr'],
        'ecnt_crpscl': ['crpscl'],
        'ecnt_crps_emp': ['crps_emp'],
        'ecnt_crpscl_emp': ['crpscl_emp'],
        'ecnt_crpss_emp': ['crpscl_emp', 'crps_emp'],
        'ecnt_crps_emp_fair': ['crps_emp_fair'],
        'ecnt_spread_md': ['spread_md'],
        'ecnt_mae': ['mae'],
        'ecnt_mae_oerr': ['mae_oerr'],
        'ecnt_n_ge_obs': [],
        'ecnt_me_ge_obs': ['me_ge_obs'],
        'ecnt_n_lt_obs': [],
        'ecnt_me_lt_obs': ['me_lt_obs'],
        'ecnt_bias_ratio': ['me_ge_obs','me_lt_obs'],

        'nbr_fbs': ['fbs'],
        'nbr_fss': ['fss'],
        'nbr_afss': ['afss'],
        'nbr_f_rare': ['f_rate'],
        'nbr_o_rare': ['o_rate'],

    }

    def _calc_stats(self, values):
        """Calculate the statistic of values for each bootstrap sample
            Args:
                values: a np.array of values we want to calculate the statistic on
                    This is actually a 2d array (matrix) of values. Each row represents
                    a bootstrap resample simulation that we wish to aggregate across.
             Returns:
                a list of calculated statistics
            Raises:
                an error

        """
        func_name = f'calculate_{self.statistic}'
        # some functions have an extra 3rd parameter that represents
        # if some data preliminary data aggregation was done
        # if this parameter is present we need to add it
        num_parameters = len(signature(globals()[func_name]).parameters)

        if values is not None and values.ndim == 2:

            # The single value case
            if num_parameters == 2:
                stat_values = [globals()[func_name](values, self.column_names)]
            else:
                stat_values = [globals()[func_name](values, self.column_names, True)]
        elif values is not None and values.ndim == 3:
            # bootstrapped case
            stat_values = []
            for row in values:
                if num_parameters == 2:
                    stat_value = [globals()[func_name](row, self.column_names)]
                else:
                    stat_value = [globals()[func_name](row, self.column_names, True)]

                stat_values.append(stat_value)

            # pool = mp.Pool(mp.cpu_count())
            # stat_values = pool.map(partial(globals()['calculate_{}'.format(stat)],
            # columns_names=columns_names), [row for row in data_for_stats])
            # pool.close()
            # pool.join()

        else:
            raise KeyError("can't calculate statistic")
        return stat_values

    def _calc_stats_derived(self, values_both_arrays):
        """Calculate the statistic of values for each derived bootstrap sample
        Args:
            values_both_arrays: a np.array of values we want to calculate the derived statistic on
                This is actually a 2d array (matrix) of values. Each row represents
                a bootstrap resample simulation that we wish to aggregate across.
                The last column contains the derived operation
                The 1st half of columns contains the 1st array data and
                the 2nd - 2nd array data
        Returns:
            a list of calculated derived statistics

        """

        if values_both_arrays is not None and values_both_arrays.ndim == 2:
            # The single value case
            num_of_columns = values_both_arrays.shape[1] - 1
            # get values for the 1st array
            values_1 = values_both_arrays[:, 0:int(num_of_columns / 2)]
            # get values for the 2nd array
            values_2 = values_both_arrays[:, int(num_of_columns / 2):num_of_columns]

            func_name = f'calculate_{self.statistic}'
            # some functions have an extra 3rd parameter that represents
            # if some data preliminary data aggregation was done
            # if this parameter is present we need to add it
            num_parameters = len(signature(globals()[func_name]).parameters)

            if num_parameters == 2:
                # calculate stat for the 1st array
                stat_values_1 = [globals()[func_name](values_1, self.column_names)]
                # calculate stat for the 2nd array
                stat_values_2 = [globals()[func_name](values_2, self.column_names)]
            else:
                # calculate stat for the 1st array
                stat_values_1 = [globals()[func_name](values_1, self.column_names, True)]
                # calculate stat for the 2nd array
                stat_values_2 = [globals()[func_name](values_2, self.column_names, True)]

            # calculate derived stat
            stat_values = calc_derived_curve_value(
                stat_values_1,
                stat_values_2,
                values_both_arrays[0, -1])
            if not isinstance(stat_values, list):
                stat_values = [stat_values]
        elif values_both_arrays is not None and values_both_arrays.ndim == 3:
            # bootstrapped case
            stat_values = []
            num_of_columns = values_both_arrays.shape[2] - 1
            for row in values_both_arrays:

                # get values for the 1st array
                values_1 = row[:, 0:int(num_of_columns / 2)]
                # get values for the 2nd array
                values_2 = row[:, int(num_of_columns / 2):num_of_columns]

                func_name = f'calculate_{self.statistic}'
                # some functions have an extra 3rd parameter that represents
                # if some data preliminary data aggregation was done
                # if this parameter is present we need to add it
                num_parameters = len(signature(globals()[func_name]).parameters)

                if num_parameters == 2:
                    # calculate stat for the 1st array
                    stat_values_1 = [globals()[func_name](values_1, self.column_names)]
                    # calculate stat for the 2nd array
                    stat_values_2 = [globals()[func_name](values_2, self.column_names)]
                else:
                    # calculate stat for the 1st array
                    stat_values_1 = [globals()[func_name](values_1, self.column_names, True)]
                    # calculate stat for the 2nd array
                    stat_values_2 = [globals()[func_name](values_2, self.column_names, True)]

                # calculate derived stat
                stat_value = calc_derived_curve_value(
                    stat_values_1,
                    stat_values_2,
                    row[0, -1])
                if not isinstance(stat_value, list):
                    stat_value = [stat_value]
                stat_values.append(stat_value)

            # pool = mp.Pool(mp.cpu_count())
            # stat_values = pool.map(partial(globals()['calculate_{}'.format(stat)],
            # columns_names=columns_names), [row for row in data_for_stats])
            # pool.close()
            # pool.join()

        else:
            raise KeyError("can't calculate statistic")
        return stat_values

    def _prepare_sl1l2_data(self, data_for_prepare):
        """Prepares sl1l2 data.
            Multiplies needed for the statistic calculation columns to the 'total'value

            Args:
                data_for_prepare: a 2d numpy array of values we want to calculate the statistic on
        """
        if self.statistic in self.STATISTIC_TO_FIELDS.keys():
            for column in self.STATISTIC_TO_FIELDS[self.statistic]:
                data_for_prepare[column] \
                    = data_for_prepare[column].values * data_for_prepare['total'].values

    def _prepare_sal1l2_data(self, data_for_prepare):
        """Prepares sal1l2 data.
            Multiplies needed for the statistic calculation columns to the 'total'value

            Args:
                data_for_prepare: a 2d numpy array of values we want to calculate the statistic on
        """
        if self.statistic in self.STATISTIC_TO_FIELDS.keys():
            for column in self.STATISTIC_TO_FIELDS[self.statistic]:
                data_for_prepare[column] \
                    = data_for_prepare[column].values * data_for_prepare['total'].values

    def _prepare_grad_data(self, data_for_prepare):
        """Prepares grad data.
            Multiplies needed for the statistic calculation columns to the 'total'value

            Args:
                data_for_prepare: a 2d numpy array of values we want to calculate the statistic on
        """
        if self.statistic in self.STATISTIC_TO_FIELDS.keys():
            for column in self.STATISTIC_TO_FIELDS[self.statistic]:
                data_for_prepare[column] \
                    = data_for_prepare[column].values * data_for_prepare['total'].values

    def _prepare_vl1l2_data(self, data_for_prepare):
        """Prepares vl1l2 data.
            Multiplies needed for the statistic calculation columns to the 'total'value

            Args:
                data_for_prepare: a 2d numpy array of values we want to calculate the statistic on
        """
        if self.statistic in self.STATISTIC_TO_FIELDS.keys():
            for column in self.STATISTIC_TO_FIELDS[self.statistic]:
                data_for_prepare[column] \
                    = data_for_prepare[column].values * data_for_prepare['total'].values

    def _prepare_val1l2_data(self, data_for_prepare):
        """Prepares val1l2 data.
            Multiplies needed for the statistic calculation columns to the 'total' value

            Args:
                data_for_prepare: a 2d numpy array of values we want to calculate the statistic on
        """
        if self.statistic in self.STATISTIC_TO_FIELDS.keys():
            for column in self.STATISTIC_TO_FIELDS[self.statistic]:
                data_for_prepare[column] \
                    = data_for_prepare[column].values * data_for_prepare['total'].values

    def _prepare_vcnt_data(self, data_for_prepare):
        """Prepares vcnt data.
            Multiplies needed for the statistic calculation columns to the 'total' value

            Args:
                data_for_prepare: a 2d numpy array of values we want to calculate the statistic on
        """
        if self.statistic in self.STATISTIC_TO_FIELDS.keys():
            for column in self.STATISTIC_TO_FIELDS[self.statistic]:
                data_for_prepare[column] \
                    = data_for_prepare[column].values * data_for_prepare['total'].values

    def _prepare_ecnt_data(self, data_for_prepare):
        """Prepares ecnt data.
            Multiplies needed for the statistic calculation columns to the 'total' value

            Args:
                data_for_prepare: a 2d numpy array of values we want to calculate the statistic on
        """
        mse = data_for_prepare['rmse'].values * data_for_prepare['rmse'].values
        mse_oerr = data_for_prepare['rmse_oerr'].values * data_for_prepare['rmse_oerr'].values
        # crps_climo = data_for_prepare['crps'].values * data_for_prepare['crps'].values

        variance = data_for_prepare['spread'].values * data_for_prepare['spread'].values
        variance_oerr = data_for_prepare['spread_oerr'].values * data_for_prepare['spread_oerr'].values
        variance_plus_oerr = data_for_prepare['spread_oerr'].values * data_for_prepare['spread_oerr'].values

        data_for_prepare['mse'] = mse * data_for_prepare['total'].values
        data_for_prepare['mse_oerr'] = mse_oerr * data_for_prepare['total'].values
        # data_for_prepare['crps_climo'] = crps_climo * data_for_prepare['total'].values

        data_for_prepare['variance'] = variance * data_for_prepare['total'].values
        data_for_prepare['variance_oerr'] = variance_oerr * data_for_prepare['total'].values
        data_for_prepare['variance_plus_oerr'] = variance_plus_oerr * data_for_prepare['total'].values

        self.column_names = data_for_prepare.columns.values

        if self.statistic in self.STATISTIC_TO_FIELDS.keys():
            for column in self.STATISTIC_TO_FIELDS[self.statistic]:
                if column == 'me_ge_obs':
                    data_for_prepare[column] \
                        = data_for_prepare[column].values * data_for_prepare['n_ge_obs'].values
                elif column == 'me_lt_obs':
                    data_for_prepare[column] \
                        = data_for_prepare[column].values * data_for_prepare['n_lt_obs'].values
                else:
                    data_for_prepare[column] \
                        = data_for_prepare[column].values * data_for_prepare['total'].values

    def _prepare_rps_data(self, data_for_prepare):
        total = data_for_prepare['total'].values
        d_rps_climo = data_for_prepare['rps'].values / (1 - data_for_prepare['rpss'].values)
        data_for_prepare['rps_climo'] = d_rps_climo * total
        data_for_prepare['rps'] = data_for_prepare['rps'].values * total
        data_for_prepare['rps_comp'] = data_for_prepare['rps_comp'].values * total
        self.column_names = data_for_prepare.columns.values

    def _prepare_ssvar_data(self, data_for_prepare):
        """Prepares ssvar data.
            Multiplies needed for the statistic calculation columns to the 'total'value

            Args:
                data_for_prepare: a 2d numpy array of values we want to calculate the statistic on
        """

        # rename bin_n column to total
        data_for_prepare.rename(columns={"total": "total_orig", "bin_n": "total"}, inplace=True)
        self.column_names = data_for_prepare.columns.values

        for column in self.STATISTIC_TO_FIELDS[self.statistic]:
            data_for_prepare[column] \
                = data_for_prepare[column].values * data_for_prepare['total'].values

    def _prepare_nbr_cnt_data(self, data_for_prepare):
        """Prepares nbrcnt data.
            Multiplies needed for the statistic calculation columns to the 'total' value

            Args:
                data_for_prepare: a 2d numpy array of values we want to calculate the statistic on
        """

        total = data_for_prepare['total'].values
        fbs = total * data_for_prepare['fbs'].values
        fss_den = (data_for_prepare['fbs'].values / (1.0 - data_for_prepare['fss'].values)) * total

        f_rate = total * data_for_prepare['f_rate'].values
        data_for_prepare['fbs'] = fbs
        data_for_prepare['fss'] = fss_den
        data_for_prepare['f_rate'] = f_rate

    def _prepare_pct_data(self, data_for_prepare):
        """Prepares pct data.
            Multiplies needed for the statistic calculation columns to the 'total'value

            Args:
                data_for_prepare: a 2d numpy array of values we want to calculate the statistic on
        """

    def _prepare_mctc_data(self, data_for_prepare):
        """Prepares mctc data.
           Nothing needs to be done

            Args:
                data_for_prepare: a 2d numpy array of values we want to calculate the statistic on
        """
        if 'ec_value' in data_for_prepare.columns:
            if not (data_for_prepare['ec_value'] == data_for_prepare['ec_value'][0]).all():
                raise ValueError('EC_VALUE is NOT constant across  MCTC lines')

    def _prepare_ctc_data(self, data_for_prepare):
        """Prepares CTC data.
            Checks if all values from ec_value column are the same and if not - throws an error

            Args:
                data_for_prepare: a 2d numpy array of values we want to calculate the statistic on
        """

        if 'ec_value' in data_for_prepare.columns:
            if not (data_for_prepare['ec_value'] == data_for_prepare['ec_value'][0]).all():
                raise ValueError('EC_VALUE is NOT constant across  CTC lines')


    def _prepare_cts_data(self, data_for_prepare):
        """Prepares cts data.
            Checks if all values from ec_value column are the same and if not - throws an error

            Args:
                data_for_prepare: a 2d numpy array of values we want to calculate the statistic on
        """

        if 'ec_value' in data_for_prepare.columns:
            if not (data_for_prepare['ec_value'] == data_for_prepare['ec_value'][0]).all():
                raise ValueError('EC_VALUE is NOT constant across  CTS lines')

    def _prepare_nbr_ctc_data(self, data_for_prepare):
        """Prepares MBR_CTC data.
            Nothing needs to be done

            Args:
                data_for_prepare: a 2d numpy array of values we want to calculate the statistic on
        """

    def _get_bootstrapped_stats_for_derived(self, series, distributions, axis="1"):
        """ Calculates aggregation derived statistic value and CI intervals if needed for input data
            Args:
                series: array of length = 3 where
                1st element - derived series title,
                    ex. 'DIFF(ENS001v3.6.1_d01 DPT FBAR-ENS001v3.6.1_d02 DPT FBAR)'
                others  - additional values like indy val and statistic
                distributions - dictionary of the series title
                    to it's BootstrapDistributionResult object

            Returns:
                BootstrapDistributionResults object

        """

        # get derived name
        derived_name = ''
        for operation in OPERATION_TO_SIGN:
            for point_component in series:
                if point_component.startswith((operation + '(', operation + ' (')):
                    derived_name = point_component
                    break
        # find all components for the 1st and 2nd series
        derived_curve_component = self.derived_name_to_values[derived_name]
        permute_for_first_series = derived_curve_component.first_component.copy()
        for series_comp in series[1:]:
            if series_comp not in permute_for_first_series:
                permute_for_first_series.append(series_comp)

        # replace first_series components group names to values
        for i, perm in enumerate(permute_for_first_series):
            if perm in self.group_to_value:
                permute_for_first_series[i] = self.group_to_value[perm]

        permute_for_second_series = derived_curve_component.second_component.copy()
        for series_comp in series[1:]:
            if series_comp not in permute_for_second_series:
                permute_for_second_series.append(series_comp)

        # replace second_series components group names to values
        for i, perm in enumerate(permute_for_second_series):
            if perm in self.group_to_value:
                permute_for_second_series[i] = self.group_to_value[perm]

        ds_1 = None
        ds_2 = None

        # for each component find  its BootstrapDistributionResult object
        for series_to_distrib_key in distributions.keys():
            if all(elem in permute_for_first_series for elem in series_to_distrib_key):
                ds_1 = distributions[series_to_distrib_key]
            if all(elem in permute_for_second_series for elem in series_to_distrib_key):
                ds_2 = distributions[series_to_distrib_key]
            if ds_1 is not None and ds_2 is not None:
                break

        # if BootstrapDistributionResult object doesn't exist
        # or the original series data size is 0 return an empty object
        if ds_1.values is None or ds_2.values is None \
                or ds_1.values.size == 0 or ds_2.values.size == 0:
            return BootstrapResults(lower_bound=None,
                                    value=None,
                                    upper_bound=None)
        # calculate the number of values in the group if the series has a group
        # it is need d for the validation
        num_diff_vals_first = 0
        num_diff_vals_second = 0
        for val in permute_for_first_series:
            size = len(val.split(GROUP_SEPARATOR))
            if size > 1:
                num_diff_vals_first = num_diff_vals_first + size
        for val in permute_for_second_series:
            size = len(val.split(GROUP_SEPARATOR))
            if size > 1:
                num_diff_vals_second = num_diff_vals_second + size
        if num_diff_vals_first == 0:
            num_diff_vals_first = 1
        if num_diff_vals_second == 0:
            num_diff_vals_second = 1

        # validate data
        if derived_curve_component.derived_operation != 'SINGLE':
            self._validate_series_cases_for_derived_operation(ds_1.values, axis, num_diff_vals_first)
            self._validate_series_cases_for_derived_operation(ds_2.values, axis, num_diff_vals_second)

        if self.params['num_iterations'] == 1 or derived_curve_component.derived_operation == 'ETB':
            # don't need bootstrapping and CI calculation -
            # calculate the derived statistic and exit

            if derived_curve_component.derived_operation == 'ETB':
                index_array = np.where(self.column_names == 'stat_value')[0]
                func_name = f'calculate_{self.statistic}'
                for row in ds_1.values:
                    stat = [globals()[func_name](row[np.newaxis, ...], self.column_names)]
                    row[index_array] = stat
                for row in ds_2.values:
                    stat = [globals()[func_name](row[np.newaxis, ...], self.column_names)]
                    row[index_array] = stat

                ds_1_value = ds_1.values[:, index_array].flatten().tolist()
                ds_2_value = ds_2.values[:, index_array].flatten().tolist()
            else:
                ds_1_value = [ds_1.value]
                ds_2_value = [ds_2.value]

            stat_val = calc_derived_curve_value(
                ds_1_value,
                ds_2_value,
                derived_curve_component.derived_operation)
            if stat_val is not None:
                results = BootstrapResults(lower_bound=None,
                                       value=round_half_up(stat_val[0], 5),
                                       upper_bound=None)
            else:
                results = BootstrapResults(lower_bound=None,
                                           value=None,
                                           upper_bound=None)
            results.set_distributions([results.value])
        else:
            # need bootstrapping and CI calculation in addition to the derived statistic

            # construct joined array with data for series 1 and 2 and operation
            operation = np.full((len(ds_1.values), 1), derived_curve_component.derived_operation)
            values_both_arrays = np.concatenate((ds_1.values, ds_2.values), axis=1)
            values_both_arrays = np.concatenate((values_both_arrays, operation), axis=1)

            try:
                # calculate a block length for the circular temporal block bootstrap if needed
                block_length = 1

                # to use circular block bootstrap or not
                is_cbb = True
                if 'circular_block_bootstrap' in self.params.keys():
                    is_cbb = parse_bool(self.params['circular_block_bootstrap'])

                if is_cbb:
                    block_length = int(math.sqrt(len(values_both_arrays)))
                results = bootstrap_and_value(
                    values_both_arrays,
                    stat_func=self._calc_stats_derived,
                    num_iterations=self.params['num_iterations'],
                    num_threads=self.params['num_threads'],
                    ci_method=self.params['method'],
                    alpha=self.params['alpha'],
                    save_data=False,
                    save_distributions=derived_curve_component.derived_operation == 'DIFF_SIG',
                    block_length=block_length)
            except KeyError as err:
                results = BootstrapResults(None, None, None)
                print(err)

        if derived_curve_component.derived_operation == 'DIFF_SIG':
            # remove None values in distributions
            distributions = [i for i in results.distributions if i is not None]
            diff_sig = None
            if distributions and results.value is not None:
                distribution_mean = np.mean(distributions)
                distribution_under_h0 = distributions - distribution_mean
                pval = np.mean(np.absolute(distribution_under_h0) <= np.absolute(results.value))
                diff_sig = perfect_score_adjustment(ds_1.value, ds_2.value, self.statistic, pval)

            results.value = diff_sig

        return results

    def _get_bootstrapped_stats(self, series_data, axis="1"):
        """ Calculates aggregation statistic value and CI intervals if needed for input data
            Args:
                series_data: pandas data frame
            Returns:
                BootstrapDistributionResults object

        """

        # if the data frame is empty - do nothing and return an empty object
        if series_data.empty:
            return BootstrapResults(lower_bound=None,
                                    value=None,
                                    upper_bound=None)
        # check if derived series are present
        has_derived_series = False
        if self.params['derived_series_' + axis]:
            has_derived_series = True

        # sort data by dates
        series_data = sort_data(series_data)
        series_data.reset_index(inplace=True, drop=True)

        if 'line_type' in self.params.keys() and self.params['line_type'] is not None and self.params['line_type'] != 'None':
            # find the function that prepares data and execute it
            func = getattr(self, f"_prepare_{self.params['line_type']}_data")
            func(series_data)

        # input data has to be in numpy format for bootstrapping
        data = series_data.to_numpy()

        if self.params['num_iterations'] == 1:
            # don't need bootstrapping and CI calculation -
            # calculate the statistic and exit
            stat_val = self._calc_stats(data)[0]
            results = BootstrapResults(lower_bound=None,
                                       value=stat_val,
                                       upper_bound=None)
            # save original data only if we need it in the future
            # for derived series calculation
            if has_derived_series:
                results.set_original_values(data)
        else:
            # need bootstrapping and CI calculation in addition to statistic
            try:
                block_length = 1
                # to use circular block bootstrap or not
                is_cbb = True
                if 'circular_block_bootstrap' in self.params.keys():
                    is_cbb = parse_bool(self.params['circular_block_bootstrap'])

                if is_cbb:
                    block_length = int(math.sqrt(len(data)))
                results = bootstrap_and_value(
                    data,
                    stat_func=self._calc_stats,
                    num_iterations=self.params['num_iterations'],
                    num_threads=self.params['num_threads'],
                    ci_method=self.params['method'],
                    save_data=has_derived_series,
                    block_length=block_length)

            except KeyError as err:
                results = BootstrapResults(None, None, None)
                print(err)
        return results

    def _validate_series_cases_for_derived_operation(self, series_data, axis="1", num_diff_vals=1):
        """ Checks if the derived curve can be calculated.
            The criteria - input array must have only unique
            (fcst_valid, fcst_lead, stat_name) cases.
            Can't calculate differences if  multiple values for one valid date/fcst_lead


            Args:
                series_data: 2d numpu array
                axis: axis of the series
                num_diff_vals: number of values in the group if the series has a group,
                    1 - otherwise
            Returns:
                 This method raises an error if this criteria is False
        """
        # find indexes of columns of interests
        fcst_lead_index = np.where(self.column_names == 'fcst_lead')[0][0]
        stat_name_index = np.where(self.column_names == 'stat_name')[0][0]
        if "fcst_valid_beg" in self.column_names:
            fcst_valid_ind = np.where(self.column_names == 'fcst_valid_beg')[0][0]
        elif "fcst_valid" in self.column_names:
            fcst_valid_ind = np.where(self.column_names == 'fcst_valid')[0][0]
        elif "fcst_init_beg" in self.column_names:
            fcst_valid_ind = \
                np.where(self.column_names == 'fcst_init_beg')[0][0]
        else:
            fcst_valid_ind = \
                np.where(self.column_names == 'fcst_init')[0][0]
        try:
            # filter columns of interest
            date_lead_stat = series_data[:, [fcst_valid_ind, fcst_lead_index, stat_name_index]]
            # find the number of unique combinations
            unique_date_size = len(set(map(tuple, date_lead_stat)))
        except TypeError as err:
            print(err)
            unique_date_size = []

        # identify rows with unique combinations
        ind = np.lexsort(
            (series_data[:, stat_name_index],
             series_data[:, fcst_lead_index], series_data[:, fcst_valid_ind]))
        series_data = series_data[ind, :]

        # the length of the frame with unique combinations should be the same
        # as the number of unique combinations calculated before

        if len(series_data) / num_diff_vals != unique_date_size \
                and self.params['list_stat_' + axis] not in self.EXEMPTED_VARS:
            raise NameError("Derived curve can't be calculated."
                            " Multiple values for one valid date/fcst_lead")

    def _init_out_frame(self, series_fields, series):
        """ Initialises the output frame and add series values to each row
            Args:
                series_fields: list of all possible series fields
                series: list of all series definitions
            Returns:
                pandas data frame
        """
        result = pd.DataFrame()
        row_number = len(series)
        # fill series variables and values
        for field_ind, field in enumerate(series_fields):
            result[field] = [row[field_ind] for row in series]

        # fill the stats  and CI values placeholders with None
        result['fcst_var'] = [None] * row_number
        result['stat_value'] = [None] * row_number
        result['stat_btcl'] = [None] * row_number
        result['stat_btcu'] = [None] * row_number
        result['nstats'] = [None] * row_number
        return result

    def _get_derived_points(self, series_val, indy_vals, axis="1"):
        """identifies and returns as an list all possible derived points values

            Args:
                series_val: dictionary of series variable to values
                indy_vals: list of independent values
            Returns: a list of all possible values for the each derived points

        """

        # for each derived series
        result = []
        for derived_serie in self.params['derived_series_' + axis]:
            # series 1 components
            ds_1 = derived_serie[0].split(' ')

            # series 2 components
            ds_2 = derived_serie[1].split(' ')
            # find a variable of the operation by comparing values in each derived series component
            series_var_vals = ()
            for ind, name in enumerate(ds_1):
                if name != ds_2[ind]:
                    series_var_vals = (name, ds_2[ind])
                    break

            series_var = list(series_val.keys())[-1]
            if len(series_var_vals) > 0:
                for var in series_val.keys():
                    if all(elem in series_val[var] for elem in series_var_vals):
                        series_var = var
                        break

            derived_val = series_val.copy()
            derived_val[series_var] = None

            for var in series_val.keys():
                if derived_val[var] is not None \
                        and intersection(derived_val[var], ds_1) \
                        == intersection(derived_val[var], ds_1):
                    derived_val[var] = intersection(derived_val[var], ds_1)

            derived_curve_name = get_derived_curve_name(derived_serie)
            derived_val[series_var] = [derived_curve_name]
            if len(indy_vals) > 0:
                derived_val[self.params['indy_var']] = indy_vals

            self.derived_name_to_values[derived_curve_name] \
                = DerivedCurveComponent(ds_1, ds_2, derived_serie[-1])
            if ds_1[-1] == ds_2[-1]:
                derived_val['stat_name'] = [ds_1[-1]]
            else:
                derived_val['stat_name'] = [ds_1[-1] + "," + ds_2[-1]]
            result.append(list(itertools.product(*derived_val.values())))

        return [y for x in result for y in x]

    def _proceed_with_axis(self, axis="1"):
        """Calculates stat values for the requested Y axis

            Args:
                axis: 1 or 2 Y axis
             Returns:
                pandas dataframe  with calculated stat values and CI

        """
        if not self.input_data.empty:
            # replace thresh_i values for reliability plot
            indy_vals = self.params['indy_vals']
            if self.params['indy_var'] == 'thresh_i' and self.params['line_type'] == 'pct':
                indy_vals_int = self.input_data['thresh_i'].tolist()
                indy_vals_int.sort()
                indy_vals_int = np.unique(indy_vals_int).tolist()
                indy_vals = list(map(str, indy_vals_int))

            # identify all possible points values by adding series values, indy values
            # and statistics and then permute them
            series_val = self.params['series_val_' + axis]
            all_fields_values = series_val.copy()
            if indy_vals:
                all_fields_values[self.params['indy_var']] = indy_vals
            all_fields_values['stat_name'] = self.params['list_stat_' + axis]
            all_points = list(itertools.product(*all_fields_values.values()))

            if self.params['derived_series_' + axis]:
                # identifies and add all possible derived points values
                all_points.extend(self._get_derived_points(series_val, indy_vals, axis))

            # init the template for output frame
            out_frame = self._init_out_frame(all_fields_values.keys(), all_points)

            point_to_distrib = {}

            # for each point
            for point_ind, point in enumerate(all_points):
                # get statistic. Use reversed because it is more likely that the stat is in the end
                for component in reversed(point):
                    if component in set(self.params['list_stat_' + axis]):
                        self.statistic = component.lower()
                        break
                is_derived = is_derived_point(point)
                if not is_derived:
                    # filter point data
                    all_filters = []
                    all_filters_pct = []
                    filters_wihtout_indy = []
                    indy_val = None
                    for field_ind, field in enumerate(all_fields_values.keys()):

                        filter_value = point[field_ind]
                        if GROUP_SEPARATOR in filter_value:
                            filter_list = re.findall(DATE_TIME_REGEX, filter_value)
                            if len(filter_list) == 0:
                                filter_list = filter_value.split(GROUP_SEPARATOR)
                        elif ";" in filter_value:
                            filter_list = filter_value.split(';')
                        else:
                            filter_list = [filter_value]
                        for i, filter_val in enumerate(filter_list):
                            if is_string_integer(filter_val):
                                filter_list[i] = int(filter_val)
                            elif is_string_strictly_float(filter_val):
                                filter_list[i] = float(filter_val)
                        if field in self.input_data.keys():
                            if field != self.params['indy_var']:  #
                                filters_wihtout_indy. \
                                    append((self.input_data[field].isin(filter_list)))
                            else:
                                indy_val = filter_value

                            all_filters.append(self.input_data[field].isin(filter_list))
                        if field in series_val.keys():
                            all_filters_pct.append((self.input_data[field].isin(filter_list)))

                    # add fcst var
                    fcst_var = None
                    if len(self.params['fcst_var_val_' + axis]) > 0:
                        fcst_var = list(self.params['fcst_var_val_' + axis].keys())[0]
                        if 'fcst_var' in self.input_data.columns:
                            all_filters.append((self.input_data['fcst_var'].isin([fcst_var])))

                    # use numpy to select the rows where any record evaluates to True
                    mask = np.array(all_filters).all(axis=0)
                    point_data = self.input_data.loc[mask]

                    if self.params['line_type'] == 'pct':
                        if all_filters_pct:
                            mask_pct = np.array(all_filters_pct).all(axis=0)
                            point_data_pct = self.input_data.loc[mask_pct]
                        else:
                            point_data_pct = self.input_data
                        # collect all columns that starts with oy_i and on_i
                        filter_oy_i = [col for col in point_data_pct if col.startswith('oy_i')]
                        filter_on_i = [col for col in point_data_pct if col.startswith('on_i')]
                        # calculate oy_total
                        oy_total = point_data_pct[filter_oy_i].values.sum()

                        # calculate T
                        sum_n_i_orig_T = point_data_pct[filter_on_i].values.sum() + oy_total

                        # calculate o_bar
                        o_bar = oy_total / sum_n_i_orig_T

                        point_data.insert(len(point_data.columns), 'T', sum_n_i_orig_T)
                        point_data.insert(len(point_data.columns), 'oy_total', oy_total)
                        point_data.insert(len(point_data.columns), 'o_bar', o_bar)

                    # aggregate point data
                    series_var_val = self.params['series_val_' + axis]
                    if any(';' in series_val for series_val in series_var_val):
                        point_data = aggregate_field_values(series_var_val, point_data, self.params['line_type'])
                    elif indy_val and ';' in indy_val:
                        # if aggregated value in indy val - add it to series values add aggregate
                        series_indy_var_val = series_var_val
                        series_indy_var_val[self.params['indy_var']] = [indy_val]
                        point_data = aggregate_field_values(series_indy_var_val, point_data,
                                                            self.params['line_type'])
                    # calculate bootstrap results
                    bootstrap_results = self._get_bootstrapped_stats(point_data, axis)
                    # save bootstrap results
                    point_to_distrib[point] = bootstrap_results
                    n_stats = len(point_data)

                else:
                    # calculate bootstrap results for the derived point
                    bootstrap_results = self._get_bootstrapped_stats_for_derived(
                        point,
                        point_to_distrib,
                        axis)
                    n_stats = 0

                # save results to the output data frame
                out_frame['fcst_var'][point_ind] = fcst_var
                out_frame['stat_value'][point_ind] = bootstrap_results.value
                out_frame['stat_btcl'][point_ind] = bootstrap_results.lower_bound
                out_frame['stat_btcu'][point_ind] = bootstrap_results.upper_bound
                out_frame['nstats'][point_ind] = n_stats

        else:
            out_frame = pd.DataFrame()
        return out_frame

    def calculate_stats_and_ci(self):
        """ Calculates aggregated statistics and confidants intervals
            ( if parameter num_iterations > 1) for each series point
            Writes output data to the file

        """

        # set random seed if present
        if self.params['random_seed'] is not None and self.params['random_seed'] != 'None':
            np.random.seed(self.params['random_seed'])

        is_event_equal = parse_bool(self.params['event_equal'])

        if self.params['line_type'] == 'pct':
            self.column_names = np.append(self.column_names, 'T')
            self.column_names = np.append(self.column_names, 'oy_total')
            self.column_names = np.append(self.column_names, 'o_bar')

        # perform grouping
        series_val = self.params['series_val_1']
        group_to_value_index = 1
        if series_val:
            for key in series_val.keys():
                for val in series_val[key]:
                    if GROUP_SEPARATOR in val:
                        new_name = 'Group_y1_' + str(group_to_value_index)
                        self.group_to_value[new_name] = val
                        group_to_value_index = group_to_value_index + 1

        series_val = self.params['series_val_2']
        if series_val:
            group_to_value_index = 1
            if series_val:
                for key in series_val.keys():
                    for val in series_val[key]:
                        if GROUP_SEPARATOR in val:
                            new_name = 'Group_y2_' + str(group_to_value_index)
                            self.group_to_value[new_name] = val
                            group_to_value_index = group_to_value_index + 1

        # perform EE if needed
        if is_event_equal:
            self.input_data = perform_event_equalization(self.params, self.input_data)

        # get results for axis1
        out_frame = self._proceed_with_axis("1")

        # get results for axis2 if needed
        if self.params['series_val_2']:
            out_frame = out_frame.append(self._proceed_with_axis("2"))

        header = True
        mode = 'w'

        if 'append_to_file' in self.params.keys() and self.params['append_to_file'] == 'True':
            header = False
            mode = 'a'

        export_csv = out_frame.to_csv(self.params['agg_stat_output'],
                                      index=None, header=header, mode=mode,
                                      sep="\t", na_rep="NA", float_format='%.'+ str(PRECISION) +'f')


if __name__ == "__main__":
    PARSER = argparse.ArgumentParser(description='List of agg_stat arguments')
    PARSER.add_argument("parameters_file", help="Path to YAML parameters file",
                        type=argparse.FileType('r'),
                        default=sys.stdin)
    ARGS = PARSER.parse_args()
    PARAMS = yaml.load(ARGS.parameters_file, Loader=yaml.FullLoader)

    AGG_STAT = AggStat(PARAMS)
    AGG_STAT.calculate_stats_and_ci()



# ============================*
# ** Copyright UCAR (c) 2020
# ** University Corporation for Atmospheric Research (UCAR)
# ** National Center for Atmospheric Research (NCAR)
# ** Research Applications Lab (RAL)
# ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
# ============================*


"""
Program Name: vcnt_statistics.py
"""
import warnings
import numpy as np

from metcalcpy.util.met_stats import calc_direction, calc_speed
from metcalcpy.util.utils import round_half_up, sum_column_data_by_name, PRECISION, get_total_values

__author__ = 'Tatiana Burek'
__version__ = '0.1.0'

from metcalcpy.util.vl1l2_statistics import calculate_vl1l2_fvar, calculate_vl1l2_ovar


def calculate_vcnt_fbar(input_data, columns_names, aggregation=False):
    """Performs calculation of VCNT_FBAR - Mean value of forecast wind speed

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated VCNT_FBAR as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        result = sum_column_data_by_name(input_data, columns_names, 'f_speed_bar') / total
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_vcnt_obar(input_data, columns_names, aggregation=False):
    """Performs calculation of VCNT_OBAR - Mean value of observed wind speed

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated VCNT_OBAR as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        result = sum_column_data_by_name(input_data, columns_names, 'o_speed_bar') / total
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_vcnt_fs_rms(input_data, columns_names, aggregation=False):
    """Performs calculation of VCNT_FS_RMS - Root mean square forecast wind speed

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated VCNT_FS_RMS as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        uvffbar = sum_column_data_by_name(input_data, columns_names, 'uvffbar') / total
        result = np.sqrt(uvffbar)
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_vcnt_os_rms(input_data, columns_names, aggregation=False):
    """Performs calculation of VCNT_OS_RMS - Root mean square observed wind speed

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated VCNT_OS_RMS as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        uvoobar = sum_column_data_by_name(input_data, columns_names, 'uvoobar') / total
        result = np.sqrt(uvoobar)
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_vcnt_msve(input_data, columns_names, aggregation=False):
    """Performs calculation of VCNT_MSVE - Mean squared length of the vector
    difference between the forecast and observed winds

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated VCNT_MSVE as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        uvffbar = sum_column_data_by_name(input_data, columns_names, 'uvffbar') / total
        uvfobar = sum_column_data_by_name(input_data, columns_names, 'uvfobar') / total
        uvoobar = sum_column_data_by_name(input_data, columns_names, 'uvoobar') / total
        mse = uvffbar - 2 * uvfobar + uvoobar
        if mse < 0:
            result = None
        else:
            result = round_half_up(mse, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_vcnt_rmsve(input_data, columns_names, aggregation=False):
    """Performs calculation of VCNT_RMSVE - Square root of Mean squared length of the vector
        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated VCNT_RMSVE as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        msve = calculate_vcnt_msve(input_data, columns_names, aggregation)
        result = np.sqrt(msve)
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_vcnt_fstdev(input_data, columns_names, aggregation=False):
    """Performs calculation of VCNT_FSTDEV - Standard deviation of the forecast wind speed
        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated VCNT_FSTDEV as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        result = np.sqrt(calculate_vl1l2_fvar(input_data, columns_names, aggregation))
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_vcnt_ostdev(input_data, columns_names, aggregation=False):
    """Performs calculation of VCNT_OSTDEV - Standard deviation of the observed wind speed
        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated VCNT_OSTDEV as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        result = np.sqrt(calculate_vl1l2_ovar(input_data, columns_names, aggregation))
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_vcnt_fdir(input_data, columns_names, aggregation=False):
    """Performs calculation of VCNT_FDIR - Direction of the average forecast wind vector
        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated VCNT_FDIR as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        ufbar = sum_column_data_by_name(input_data, columns_names, 'ufbar') / total
        vfbar = sum_column_data_by_name(input_data, columns_names, 'vfbar') / total
        fdir = calc_direction(-ufbar, -vfbar)
        result = round_half_up(fdir, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_vcnt_odir(input_data, columns_names, aggregation=False):
    """Performs calculation of VCNT_ODIR - Direction of the average observed wind vector
        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated VCNT_ODIR as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        uobar = sum_column_data_by_name(input_data, columns_names, 'uobar') / total
        vobar = sum_column_data_by_name(input_data, columns_names, 'vobar') / total
        odir = calc_direction(-uobar, -vobar)
        result = round_half_up(odir, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_vcnt_fbar_speed(input_data, columns_names, aggregation=False):
    """Performs calculation of VCNT_FBAR_SPEED - Length (speed) of the average forecast wind vector
        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated VCNT_FBAR_SPEED as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        ufbar = sum_column_data_by_name(input_data, columns_names, 'ufbar') / total
        vfbar = sum_column_data_by_name(input_data, columns_names, 'vfbar') / total
        fspd = calc_speed(ufbar, vfbar)
        result = round_half_up(fspd, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_vcnt_obar_speed(input_data, columns_names, aggregation=False):
    """Performs calculation of VCNT_OBAR_SPEED - Length (speed) of the average observed wind vector
        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated VCNT_OBAR_SPEED as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        uobar = sum_column_data_by_name(input_data, columns_names, 'uobar') / total
        vobar = sum_column_data_by_name(input_data, columns_names, 'vobar') / total
        fspd = calc_speed(uobar, vobar)
        result = round_half_up(fspd, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_vcnt_vdiff_speed(input_data, columns_names, aggregation=False):
    """Performs calculation of VCNT_VDIFF_SPEED - Length (speed)  of the vector deference between
    the average forecast and average observed wind vectors

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated VCNT_VDIFF_SPEED as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        ufbar = sum_column_data_by_name(input_data, columns_names, 'ufbar') / total
        uobar = sum_column_data_by_name(input_data, columns_names, 'uobar') / total
        vfbar = sum_column_data_by_name(input_data, columns_names, 'vfbar') / total
        vobar = sum_column_data_by_name(input_data, columns_names, 'vobar') / total
        vdiff_spd = calc_speed(ufbar - uobar, vfbar - vobar)
        result = round_half_up(vdiff_spd, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_vcnt_vdiff_dir(input_data, columns_names, aggregation=False):
    """Performs calculation of VCNT_VDIFF_DIR - Direction of the vector deference between
    the average forecast and average wind vector

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated VCNT_VDIFF_DIR as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation)
        ufbar = sum_column_data_by_name(input_data, columns_names, 'ufbar') / total
        uobar = sum_column_data_by_name(input_data, columns_names, 'uobar') / total
        vfbar = sum_column_data_by_name(input_data, columns_names, 'vfbar') / total
        vobar = sum_column_data_by_name(input_data, columns_names, 'vobar') / total
        vdiff_dir = calc_direction(-(ufbar - uobar), -(vfbar - vobar))
        result = round_half_up(vdiff_dir, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_vcnt_speed_err(input_data, columns_names, aggregation=False):
    """Performs calculation of VCNT_SPEED_ERR - Deference between
        the length of the average forecast wind vector
     and the average observed wind vector (in the sense F - O)

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated VCNT_SPEED_ERR as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        speed_bias = calculate_vcnt_fbar_speed(input_data, columns_names, aggregation) \
                     - calculate_vcnt_obar_speed(input_data, columns_names, aggregation)
        result = round_half_up(speed_bias, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_vcnt_speed_abserr(input_data, columns_names, aggregation=False):
    """Performs calculation of VCNT_SPEED_ABSERR - Absolute value of diference between the length
     of the average forecast wind vector
     and the average observed wind vector (in the sense F - O)

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated VCNT_SPEED_ABSERR as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        spd_abserr = abs(calculate_vcnt_speed_err(input_data, columns_names, aggregation))
        result = round_half_up(spd_abserr, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_vcnt_dir_err(input_data, columns_names, aggregation=False):
    """Performs calculation of VCNT_DIR_ERR - Signed angle between the directions
        of the average forecast and observed wind vectors.
        Positive if the forecast wind vector is counter clockwise from the observed wind vector

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated VCNT_DIR_ERR as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        f_len = calculate_vcnt_fbar_speed(input_data, columns_names, aggregation)
        total = get_total_values(input_data, columns_names, aggregation)
        ufbar = sum_column_data_by_name(input_data, columns_names, 'ufbar') / total
        vfbar = sum_column_data_by_name(input_data, columns_names, 'vfbar') / total
        uf = ufbar / f_len
        vf = vfbar / f_len

        o_len = calculate_vcnt_obar_speed(input_data, columns_names)
        uobar = sum_column_data_by_name(input_data, columns_names, 'uobar') / total
        vobar = sum_column_data_by_name(input_data, columns_names, 'vobar') / total
        uo = uobar / o_len
        vo = vobar / o_len

        a = vf * uo - uf * vo
        b = uf * uo + vf * vo

        dir_err = calc_direction(a, b)

        result = round_half_up(dir_err, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_vcnt_dir_abser(input_data, columns_names, aggregation=False):
    """Performs calculation of VCNT_DIR_ABSERR - Absolute value of
        signed angle between the directions of the average forecast
        and observed wind vectors. Positive if the forecast wind vector
        is counter clockwise from the observed wind vector

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array
            aggregation: if the aggregation on fields was performed

        Returns:
            calculated VCNT_DIR_ABSERR as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        ang_btw = abs(calculate_vcnt_dir_err(input_data, columns_names, aggregation))
        result = round_half_up(ang_btw, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_vcnt_anom_corr(input_data, columns_names, aggregation=False):
    warnings.filterwarnings('error')
    try:
        total = get_total_values(input_data, columns_names, aggregation) # n
        fa_speed_bar = sum_column_data_by_name(input_data, columns_names, 'fa_speed_bar')   # f
        oa_speed_bar = sum_column_data_by_name(input_data, columns_names, 'oa_speed_bar')   # o
        uvffabar = sum_column_data_by_name(input_data, columns_names, 'uvffabar')   # ff
        uvfoabar = sum_column_data_by_name(input_data, columns_names, 'uvfoabar')  # fo
        uvooabar = sum_column_data_by_name(input_data, columns_names, 'uvooabar')   # oo

        v = (total * uvffabar - fa_speed_bar * fa_speed_bar) * (total * uvooabar - oa_speed_bar * oa_speed_bar)
        result = ((total * uvfoabar) - (fa_speed_bar * oa_speed_bar)) / np.sqrt(v)

        # Check the computed range
        if result > 1:
            result = 1.0
        elif result < -1:
            result = -1.0

    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_vcnt_anom_corr_uncntr(input_data, columns_names, aggregation=False):
    warnings.filterwarnings('error')
    try:
        uvffabar = sum_column_data_by_name(input_data, columns_names, 'uvffabar')   # ff
        uvooabar = sum_column_data_by_name(input_data, columns_names, 'uvooabar')   # oo
        uvfoabar = sum_column_data_by_name(input_data, columns_names, 'uvfoabar')   # fo

        v = uvffabar * uvooabar
        result = uvfoabar / np.sqrt(v)

        # Check the computed range
        if result > 1:
            result = 1.0
        elif result < -1:
            result = -1.0

    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result



# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
"""
Program Name: event_equalize_against_values.py
"""

import pandas as pd

__author__ = 'Tatiana Burek'
__version__ = '0.1.0'


def event_equalize_against_values(series_data, input_unique_cases):
    """Performs event equalisation.

    event_equalize_against_values assumes that the input series_data contains data
    indexed by fcst_valid, series values and the independent variable values.
    It builds a new data frame which contains the same
    data except for records that don't have corresponding fcst_valid
    and fcst_lead values from ee_stats_equalize

    Args:
        series_data: data frame containing the records to equalize, including fcst_valid_beg, series
                  values and independent variable values

        input_unique_cases: unique cases to equalize against
    Returns:
        A data frame that contains equalized records or empty frame
    """

    warning_remove = "WARNING: event equalization removed {} rows"

    column_names = list(series_data)

    if 'fcst_valid' in column_names:
        # always use fcst_valid for equalization
        # create a unique member to use for equalization
        series_data.insert(len(series_data.columns), 'equalize',
                           series_data['fcst_valid'].astype(str)
                           + ' '
                           + series_data['fcst_lead'].astype(str))
    else:
        print("WARNING: eventEqualize() did not run due to lack of valid time field")
        return pd.DataFrame()

    # create an equalized set of data for the minimal list of dates based on the input cases
    data_for_unique_cases = series_data[(series_data['equalize'].isin(input_unique_cases))]
    n_row_cases = len(data_for_unique_cases)
    if n_row_cases == 0:
        print(" WARNING: discarding all members")
        return pd.DataFrame()

    n_row_ = len(series_data)
    if n_row_cases != n_row_:
        print(warning_remove.format(n_row_ - n_row_cases))

    # remove 'equalize' column
    data_for_unique_cases = data_for_unique_cases.drop(['equalize'], axis=1)
    return data_for_unique_cases



# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
"""
Program Name: agg_stat_bootstrap.py

How to use:
 - Call from other Python function
        AAGG_STAT_BOOTSTRAP = AggStatBootstrap(PARAMS)
        AGG_STAT_BOOTSTRAP.calculate_values()
        where PARAMS – a dictionary with data description parameters including
        location of input and output data.
        The structure is similar to Rscript template

 - Run as a stand-alone script
        python agg_stat_bootstrap.py <parameters_file>
        where - <parameters_file> is YAML file with parameters
        and environment variable should be set to PYTHONPATH=<path_to_METcalcpy>

 - Run from Java
        proc = Runtime.getRuntime().exec(
                “python agg_stat_bootstrap.py <parameters_file>”,
                new String[]{”PYTHONPATH=<path_to_METcalcpy>”},
                new File(System.getProperty("user.home")));

"""
import argparse
import itertools
import sys

import yaml
import pandas as pd
import numpy as np

from metcalcpy import GROUP_SEPARATOR
from metcalcpy.agg_stat_eqz import AggStatEventEqz
from metcalcpy.bootstrap import  bootstrap_and_value_mode, BootstrapResults
from metcalcpy.util.mode_ratio_statistics import *
from metcalcpy.util.mode_arearat_statistics import *
from metcalcpy.util.mode_2d_arearat_statistics import *
from metcalcpy.util.mode_2d_ratio_statistics import *
from metcalcpy.util.mode_3d_volrat_statistics import *
from metcalcpy.util.mode_3d_ratio_statistics import *
from metcalcpy.util.utils import is_string_integer, parse_bool, sort_data, is_string_strictly_float


class AggStatBootstrap:
    """A class that performs aggregation statistic logic fot MODE and MTD ratio statistics on input data frame.
        All parameters including data description and location is in the parameters dictionary
        Usage:
            initialise this call with the parameters dictionary and than
            call calculate_value_and_ci method
            This method will crate and save to the file aggregation statistics
                agg_stat_boot = AggStatBootstrap(params)
                agg_stat_boot.calculate_value_and_ci()
           """

    def __init__(self, in_params):
        """Initialises the class by saving input parameters and reading data from file

            Args:
                in_params - input parameters as a dictionary
        """

        self.statistic = None
        self.derived_name_to_values = {}
        self.params = in_params

        self.input_data = pd.read_csv(
            self.params['agg_stat_input'],
            header=[0],
            sep='\t'
        )
        self.column_names = self.input_data.columns.values
        self.group_to_value = {}
        self.series_data = None

    def _init_out_frame(self, series_fields, series):
        """ Initialises the output frame and add series values to each row
            Args:
                series_fields: list of all possible series fields
                series: list of all series definitions
            Returns:
                pandas data frame
        """
        result = pd.DataFrame()
        row_number = len(series)
        # fill series variables and values
        for field_ind, field in enumerate(series_fields):
            result[field] = [row[field_ind] for row in series]

        # fill the stats  and CI values placeholders with None
        result['fcst_var'] = [None] * row_number
        result['stat_value'] = [None] * row_number
        result['stat_btcl'] = [None] * row_number
        result['stat_btcu'] = [None] * row_number
        result['nstats'] = [None] * row_number
        return result

    def _proceed_with_axis(self, axis="1"):
        if not self.input_data.empty:
            # identify all possible points values by adding series values, indy values
            # and statistics and then permute them
            indy_vals = self.params['indy_vals']
            series_val = self.params['series_val_' + axis]
            all_fields_values = series_val.copy()
            all_fields_values[self.params['indy_var']] = indy_vals
            all_fields_values['stat_name'] = self.params['list_stat_' + axis]
            all_points = list(itertools.product(*all_fields_values.values()))
            fcst_var = None
            if len(self.params['fcst_var_val_' + axis]) > 0 and 'fcst_var' in self.input_data.columns:
                fcst_var = list(self.params['fcst_var_val_' + axis].keys())[0]

            cases = []
            out_frame = self._init_out_frame(all_fields_values.keys(), all_points)
            point_to_distrib = {}

            # run the bootstrap flow for each independent variable value
            for indy_val in indy_vals:
                # extract the records for the current indy value
                if is_string_integer(indy_val):
                    filtered_by_indy_data = \
                        self.input_data[self.input_data[self.params['indy_var']] == int(indy_val)]
                elif is_string_strictly_float(indy_val):
                    filtered_by_indy_data = \
                        self.input_data[self.input_data[self.params['indy_var']] == float(indy_val)]
                else:
                    filtered_by_indy_data = \
                        self.input_data[self.input_data[self.params['indy_var']] == indy_val]
                # and statistics and then permute them
                all_fields_values = series_val.copy()

                all_points = list(itertools.product(*all_fields_values.values()))

                for point in all_points:
                    all_filters = []
                    for field_ind, field in enumerate(all_fields_values.keys()):
                        filter_value = point[field_ind]
                        if GROUP_SEPARATOR in filter_value:
                            filter_list = filter_value.split(GROUP_SEPARATOR)
                        elif ';' in filter_value:
                            filter_list = filter_value.split(';')
                        else:
                            filter_list = [filter_value]
                        for i, filter_val in enumerate(filter_list):
                            if is_string_integer(filter_val):
                                filter_list[i] = int(filter_val)
                            elif is_string_strictly_float(filter_val):
                                filter_list[i] = float(filter_val)

                        all_filters.append((filtered_by_indy_data[field].isin(filter_list)))

                    # add fcst var
                    if fcst_var is not None:
                        all_filters.append((filtered_by_indy_data['fcst_var'].isin([fcst_var])))

                    # use numpy to select the rows where any record evaluates to True
                    mask = np.array(all_filters).all(axis=0)
                    point_data = filtered_by_indy_data.loc[mask]

                    # build a list of cases to sample
                    fcst_valid = point_data.loc[:, 'fcst_valid'].astype(str)
                    indy_var = point_data.loc[:, self.params['indy_var']].astype(str)
                    case_cur = np.unique(fcst_valid + '#' + indy_var)
                    cases = np.sort(np.unique(np.append(cases, case_cur)))
                    cases = np.reshape(cases, (cases.shape[0], 1))
                # calculate bootstrap for cases
                for stat_upper in self.params['list_stat_' + axis]:
                    self.statistic = stat_upper.lower()
                    for point in all_points:
                        all_filters = []
                        out_frame_filter = []
                        for field_ind, field in enumerate(all_fields_values.keys()):
                            filter_value = point[field_ind]
                            if ':' in filter_value:
                                filter_list = filter_value.split(':')
                            elif ';' in filter_value:
                                filter_list = filter_value.split(';')
                            else:
                                filter_list = [filter_value]
                            for i, filter_val in enumerate(filter_list):
                                if is_string_integer(filter_val):
                                    filter_list[i] = int(filter_val)
                                elif is_string_strictly_float(filter_val):
                                    filter_list[i] = float(filter_val)

                            all_filters.append((filtered_by_indy_data[field].isin(filter_list)))
                            out_frame_filter.append((out_frame[field].isin(filter_list)))
                        # use numpy to select the rows where any record evaluates to True
                        mask = np.array(all_filters).all(axis=0)
                        mask_out_frame = np.array(out_frame_filter).all(axis=0)
                        point_data = filtered_by_indy_data.loc[mask]
                        bootstrap_results = self._get_bootstrapped_stats(point_data, cases)
                        # save bootstrap results
                        point_to_distrib[point] = bootstrap_results
                        n_stats = len(point_data)

                        # find an index of this point in the output data frame
                        rows_with_mask = out_frame.loc[mask_out_frame]
                        rows_with_mask_indy_var = \
                            rows_with_mask[(rows_with_mask[self.params['indy_var']] == indy_val)]
                        index = rows_with_mask_indy_var.index[0]

                        # save results to the output data frame
                        out_frame['fcst_var'][index] = fcst_var
                        out_frame['stat_value'][index] = bootstrap_results.value
                        out_frame['stat_btcl'][index] = bootstrap_results.lower_bound
                        out_frame['stat_btcu'][index] = bootstrap_results.upper_bound
                        out_frame['nstats'][index] = n_stats
        else:
            out_frame = pd.DataFrame()
        return out_frame

    def _get_bootstrapped_stats(self, series_data, cases):
        self.series_data = sort_data(series_data)
        if self.params['num_iterations'] == 1:
            stat_val = self._calc_stats(cases)[0]
            results = BootstrapResults(lower_bound=None,
                                                   value=stat_val,
                                                   upper_bound=None)
        else:
            # need bootstrapping and CI calculation in addition to statistic
            try:
                results = bootstrap_and_value_mode(
                    self.series_data,
                    cases,
                    stat_func=self._calc_stats,
                    num_iterations=self.params['num_iterations'],
                    num_threads=self.params['num_threads'],
                    ci_method=self.params['method'])

            except KeyError as err:
                results = BootstrapResults(None, None, None)
                print(err)
        return results

    def _calc_stats(self, cases):
        """Calculate the statistic of values for each bootstrap sample
            Args:
                cases: a np.array of values we want to calculate the statistic on
                    This is actually a 2d array (matrix) of values. Each row represents
                    a bootstrap resample simulation that we wish to aggregate across.
             Returns:
                a list of calculated statistics
            Raises:
                an error

        """
        func_name = f'calculate_{self.statistic}'
        if cases is not None and cases.ndim == 2:
            # The single value case

            # build a data frame with the sampled data
            data_cases = np.asarray(self.series_data['case'])
            flat_cases = cases.flatten()
            values = self.series_data[np.in1d(data_cases, flat_cases)].to_numpy()
            stat_values = [globals()[func_name](values, self.column_names)]
        elif cases is not None and cases.ndim == 3:
            # bootstrapped case
            stat_values = []
            for row in cases:
                values_ind = self.series_data['case'].isin(row.flatten())
                values = self.series_data[values_ind]
                stat_values.append([globals()[func_name](values, self.column_names)])
        else:
            raise KeyError("can't calculate statistic")
        return stat_values

    def calculate_values(self):
        """ Performs EE if needed followed by  aggregation statistic logic
            Writes output data to the file
        """
        if not self.input_data.empty:
            if self.params['random_seed'] is not None and self.params['random_seed'] != 'None':
                np.random.seed(self.params['random_seed'])

            # perform EE if needed
            is_event_equal = parse_bool(self.params['event_equal'])
            if is_event_equal:
                self._perform_event_equalization()

            # build the case information for each record
            fcst_valid = self.input_data.loc[:, 'fcst_valid'].astype(str)
            indy_var = self.input_data.loc[:, self.params['indy_var']].astype(str)
            self.input_data['case'] = fcst_valid + '#' + indy_var

            # get results for axis1
            out_frame = self._proceed_with_axis("1")
            if self.params['series_val_2']:
                out_frame = out_frame.append(self._proceed_with_axis("2"))

        else:
            out_frame = pd.DataFrame()

        header = True
        mode = 'w'

        export_csv = out_frame.to_csv(self.params['agg_stat_output'],
                                      index=None, header=header, mode=mode,
                                      sep="\t", na_rep="NA")

    def _perform_event_equalization(self):
        """ Performs event equalisation on input data
        """
        agg_stat_event_eqz = AggStatEventEqz(self.params)

        self.input_data = agg_stat_event_eqz.calculate_values()
        file_name = self.params['agg_stat_input'].replace("agg_stat_bootstrap", "dataAfterEq")
        self.input_data.to_csv(file_name, index=None, header=True, mode='w', sep="\t", na_rep="NA")


if __name__ == "__main__":
    PARSER = argparse.ArgumentParser(description='List of agg_stat_bootstrap arguments')
    PARSER.add_argument("parameters_file", help="Path to YAML parameters file",
                        type=argparse.FileType('r'),
                        default=sys.stdin)
    ARGS = PARSER.parse_args()
    PARAMS = yaml.load(ARGS.parameters_file, Loader=yaml.FullLoader)

    AGG_STAT_BOOTSTRAP = AggStatBootstrap(PARAMS)
    AGG_STAT_BOOTSTRAP.calculate_values()
